# Copyright (c) - 2015, Alex Meade
# Copyright (c) - 2015, Yogesh Kshirsagar
# Copyright (c) - 2015, Michael Price
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import json

import mock

from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.volume import configuration as conf
from cinder.volume.drivers.netapp.eseries import utils
import cinder.volume.drivers.netapp.options as na_opts
import cinder.volume.drivers.netapp.utils as na_utils

FAKE_CINDER_VOLUME = {
    'id': fake.VOLUME_ID,
    'size': 1,
    'volume_name': 'lun1',
    'host': 'hostname@backend#DDP',
    'os_type': 'linux',
    'provider_location': 'lun1',
    'name_id': fake.VOLUME2_ID,
    'provider_auth': 'provider a b',
    'project_id': fake.PROJECT_ID,
    'display_name': None,
    'display_description': 'lun1',
    'volume_type_id': None,
    'migration_status': None,
    'attach_status': fields.VolumeAttachStatus.DETACHED
}

FAKE_CINDER_SNAPSHOT = {
    'id': fake.SNAPSHOT_ID,
    'volume': FAKE_CINDER_VOLUME,
    'provider_id': '3400000060080E500023BB3400631F335294A5A8',
}

FAKE_CINDER_CG = {
    'id': fake.CONSISTENCY_GROUP_ID,
}

FAKE_CINDER_CG_SNAPSHOT = {
    'id': fake.CGSNAPSHOT_ID,
    'consistencygroup_id': FAKE_CINDER_CG['id'],
}

MULTIATTACH_HOST_GROUP = {
    'clusterRef': '8500000060080E500023C7340036035F515B78FC',
    'label': utils.MULTI_ATTACH_HOST_GROUP_NAME,
}

FOREIGN_HOST_GROUP = {
    'clusterRef': '8500000060080E500023C7340036035F515B78FD',
    'label': 'FOREIGN HOST GROUP',
}

HOST_GROUPS = [MULTIATTACH_HOST_GROUP, FOREIGN_HOST_GROUP]

SSC_POOLS = [
    {
        "poolId": "0400000060080E5000290D8000009C9955828DD2",
        "name": "DDP",
        "pool": {
            "sequenceNum": 2,
            "offline": False,
            "raidLevel": "raidDiskPool",
            "worldWideName": "60080E5000290D8000009C9955828DD2",
            "volumeGroupRef": "0400000060080E5000290D8000009C9955828DD2",
            "reserved1": "000000000000000000000000",
            "reserved2": "",
            "trayLossProtection": False,
            "label": "DDP",
            "state": "complete",
            "spindleSpeedMatch": True,
            "spindleSpeed": 7200,
            "isInaccessible": False,
            "securityType": "none",
            "drawerLossProtection": True,
            "protectionInformationCapable": False,
            "protectionInformationCapabilities": {
                "protectionInformationCapable": True,
                "protectionType": "type2Protection"
            },
            "volumeGroupData": {
                "type": "diskPool",
                "diskPoolData": {
                    "reconstructionReservedDriveCount": 1,
                    "reconstructionReservedAmt": "2992518463488",
                    "reconstructionReservedDriveCountCurrent": 1,
                    "poolUtilizationWarningThreshold": 100,
                    "poolUtilizationCriticalThreshold": 100,
                    "poolUtilizationState": "utilizationOptimal",
                    "unusableCapacity": "0",
                    "degradedReconstructPriority": "high",
                    "criticalReconstructPriority": "highest",
                    "backgroundOperationPriority": "low",
                    "allocGranularity": "4294967296"
                }
            },
            "usage": "standard",
            "driveBlockFormat": "allNative",
            "reservedSpaceAllocated": True,
            "usedSpace": "13653701033984",
            "totalRaidedSpace": "23459111370752",
            "extents": [
                {
                    "sectorOffset": "0",
                    "rawCapacity": "9805410336768",
                    "raidLevel": "raidDiskPool",
                    "volumeGroupRef":
                        "0400000060080E5000290D8000009C9955828DD2",
                    "freeExtentRef":
                        "0301000060080E5000290D8000009C9955828DD2",
                    "reserved1": "000000000000000000000000",
                    "reserved2": ""
                }
            ],
            "largestFreeExtentSize": "9805410336768",
            "raidStatus": "optimal",
            "freeSpace": "9805410336768",
            "drivePhysicalType": "sas",
            "driveMediaType": "hdd",
            "normalizedSpindleSpeed": "spindleSpeed7200",
            "id": "0400000060080E5000290D8000009C9955828DD2",
            "diskPool": True,
            "name": "DDP"
        },
        "flashCacheCapable": True,
        "dataAssuranceCapable": True,
        "encrypted": False,
        "thinProvisioningCapable": True,
        "spindleSpeed": "spindleSpeed7200",
        "raidLevel": "raidDiskPool",
        "availableFreeExtentCapacities": [
            "9805410336768"
        ]
    },
    {
        "poolId": "0400000060080E5000290D8000009CBA55828E96",
        "name": "pool_raid1",
        "pool": {
            "sequenceNum": 6,
            "offline": False,
            "raidLevel": "raid1",
            "worldWideName": "60080E5000290D8000009CBA55828E96",
            "volumeGroupRef": "0400000060080E5000290D8000009CBA55828E96",
            "reserved1": "000000000000000000000000",
            "reserved2": "",
            "trayLossProtection": False,
            "label": "pool_raid1",
            "state": "complete",
            "spindleSpeedMatch": True,
            "spindleSpeed": 10000,
            "isInaccessible": False,
            "securityType": "none",
            "drawerLossProtection": True,
            "protectionInformationCapable": False,
            "protectionInformationCapabilities": {
                "protectionInformationCapable": True,
                "protectionType": "type2Protection"
            },
            "volumeGroupData": {
                "type": "unknown",
                "diskPoolData": None
            },
            "usage": "standard",
            "driveBlockFormat": "allNative",
            "reservedSpaceAllocated": True,
            "usedSpace": "2978559819776",
            "totalRaidedSpace": "6662444097536",
            "extents": [
                {
                    "sectorOffset": "387891200",
                    "rawCapacity": "3683884277760",
                    "raidLevel": "raid1",
                    "volumeGroupRef":
                        "0400000060080E5000290D8000009CBA55828E96",
                    "freeExtentRef":
                        "030000B360080E5000290D8000009CBA55828E96",
                    "reserved1": "000000000000000000000000",
                    "reserved2": ""
                }
            ],
            "largestFreeExtentSize": "3683884277760",
            "raidStatus": "optimal",
            "freeSpace": "3683884277760",
            "drivePhysicalType": "sas",
            "driveMediaType": "hdd",
            "normalizedSpindleSpeed": "spindleSpeed10k",
            "id": "0400000060080E5000290D8000009CBA55828E96",
            "diskPool": False,
            "name": "pool_raid1"
        },
        "flashCacheCapable": False,
        "dataAssuranceCapable": True,
        "encrypted": False,
        "thinProvisioningCapable": False,
        "spindleSpeed": "spindleSpeed10k",
        "raidLevel": "raid1",
        "availableFreeExtentCapacities": [
            "3683884277760"
        ]
    },
    {
        "poolId": "0400000060080E5000290D8000009CAB55828E51",
        "name": "pool_raid6",
        "pool": {
            "sequenceNum": 3,
            "offline": False,
            "raidLevel": "raid6",
            "worldWideName": "60080E5000290D8000009CAB55828E51",
            "volumeGroupRef": "0400000060080E5000290D8000009CAB55828E51",
            "reserved1": "000000000000000000000000",
            "reserved2": "",
            "trayLossProtection": False,
            "label": "pool_raid6",
            "state": "complete",
            "spindleSpeedMatch": True,
            "spindleSpeed": 15000,
            "isInaccessible": False,
            "securityType": "enabled",
            "drawerLossProtection": False,
            "protectionInformationCapable": False,
            "protectionInformationCapabilities": {
                "protectionInformationCapable": True,
                "protectionType": "type2Protection"
            },
            "volumeGroupData": {
                "type": "unknown",
                "diskPoolData": None
            },
            "usage": "standard",
            "driveBlockFormat": "allNative",
            "reservedSpaceAllocated": True,
            "usedSpace": "16413217521664",
            "totalRaidedSpace": "16637410312192",
            "extents": [
                {
                    "sectorOffset": "1144950784",
                    "rawCapacity": "224192790528",
                    "raidLevel": "raid6",
                    "volumeGroupRef":
                        "0400000060080E5000290D8000009CAB55828E51",
                    "freeExtentRef":
                        "0300005960080E5000290D8000009CAB55828E51",
                    "reserved1": "000000000000000000000000",
                    "reserved2": ""
                }
            ],
            "largestFreeExtentSize": "224192790528",
            "raidStatus": "optimal",
            "freeSpace": "224192790528",
            "drivePhysicalType": "sas",
            "driveMediaType": "hdd",
            "normalizedSpindleSpeed": "spindleSpeed15k",
            "id": "0400000060080E5000290D8000009CAB55828E51",
            "diskPool": False,
            "name": "pool_raid6"
        },
        "flashCacheCapable": False,
        "dataAssuranceCapable": True,
        "encrypted": True,
        "thinProvisioningCapable": False,
        "spindleSpeed": "spindleSpeed15k",
        "raidLevel": "raid6",
        "availableFreeExtentCapacities": [
            "224192790528"
        ]
    }
]

STORAGE_POOLS = [ssc_pool['pool'] for ssc_pool in SSC_POOLS]

VOLUMES = [
    {
        "offline": False,
        "extremeProtection": False,
        "volumeHandle": 2,
        "raidLevel": "raid0",
        "sectorOffset": "0",
        "worldWideName": "60080E50002998A00000945355C37C19",
        "label": "1",
        "blkSize": 512,
        "capacity": "10737418240",
        "reconPriority": 1,
        "segmentSize": 131072,
        "action": "initializing",
        "cache": {
            "cwob": False,
            "enterpriseCacheDump": False,
            "mirrorActive": True,
            "mirrorEnable": True,
            "readCacheActive": True,
            "readCacheEnable": True,
            "writeCacheActive": True,
            "writeCacheEnable": True,
            "cacheFlushModifier": "flush10Sec",
            "readAheadMultiplier": 1
        },
        "mediaScan": {
            "enable": False,
            "parityValidationEnable": False
        },
        "volumeRef": "0200000060080E50002998A00000945355C37C19",
        "status": "optimal",
        "volumeGroupRef": "fakevolgroupref",
        "currentManager": "070000000000000000000001",
        "preferredManager": "070000000000000000000001",
        "perms": {
            "mapToLUN": True,
            "snapShot": True,
            "format": True,
            "reconfigure": True,
            "mirrorPrimary": True,
            "mirrorSecondary": True,
            "copySource": True,
            "copyTarget": True,
            "readable": True,
            "writable": True,
            "rollback": True,
            "mirrorSync": True,
            "newImage": True,
            "allowDVE": True,
            "allowDSS": True,
            "concatVolumeMember": True,
            "flashReadCache": True,
            "asyncMirrorPrimary": True,
            "asyncMirrorSecondary": True,
            "pitGroup": True,
            "cacheParametersChangeable": True,
            "allowThinManualExpansion": False,
            "allowThinGrowthParametersChange": False,
            "allowVaulting": False,
            "allowRestore": False
        },
        "mgmtClientAttribute": 0,
        "dssPreallocEnabled": True,
        "dssMaxSegmentSize": 2097152,
        "preReadRedundancyCheckEnabled": False,
        "protectionInformationCapable": False,
        "protectionType": "type1Protection",
        "applicationTagOwned": False,
        "untrustworthy": 0,
        "volumeUse": "standardVolume",
        "volumeFull": False,
        "volumeCopyTarget": False,
        "volumeCopySource": False,
        "pitBaseVolume": False,
        "asyncMirrorTarget": False,
        "asyncMirrorSource": False,
        "remoteMirrorSource": False,
        "remoteMirrorTarget": False,
        "diskPool": False,
        "flashCached": False,
        "increasingBy": "0",
        "metadata": [],
        "dataAssurance": True,
        "name": "1",
        "id": "0200000060080E50002998A00000945355C37C19",
        "wwn": "60080E50002998A00000945355C37C19",
        "objectType": "volume",
        "mapped": False,
        "preferredControllerId": "070000000000000000000001",
        "totalSizeInBytes": "10737418240",
        "onlineVolumeCopy": False,
        "listOfMappings": [],
        "currentControllerId": "070000000000000000000001",
        "cacheSettings": {
            "cwob": False,
            "enterpriseCacheDump": False,
            "mirrorActive": True,
            "mirrorEnable": True,
            "readCacheActive": True,
            "readCacheEnable": True,
            "writeCacheActive": True,
            "writeCacheEnable": True,
            "cacheFlushModifier": "flush10Sec",
            "readAheadMultiplier": 1
        },
        "thinProvisioned": False
    },
    {
        "volumeHandle": 16385,
        "worldWideName": "60080E500029347000001D7B55C3791E",
        "label": "2",
        "allocationGranularity": 128,
        "capacity": "53687091200",
        "reconPriority": 1,
        "volumeRef": "3A00000060080E500029347000001D7B55C3791E",
        "status": "optimal",
        "repositoryRef": "3600000060080E500029347000001D7955C3791D",
        "currentManager": "070000000000000000000002",
        "preferredManager": "070000000000000000000002",
        "perms": {
            "mapToLUN": True,
            "snapShot": False,
            "format": True,
            "reconfigure": False,
            "mirrorPrimary": False,
            "mirrorSecondary": False,
            "copySource": True,
            "copyTarget": False,
            "readable": True,
            "writable": True,
            "rollback": True,
            "mirrorSync": True,
            "newImage": True,
            "allowDVE": True,
            "allowDSS": True,
            "concatVolumeMember": False,
            "flashReadCache": True,
            "asyncMirrorPrimary": True,
            "asyncMirrorSecondary": True,
            "pitGroup": True,
            "cacheParametersChangeable": True,
            "allowThinManualExpansion": False,
            "allowThinGrowthParametersChange": False,
            "allowVaulting": False,
            "allowRestore": False
        },
        "mgmtClientAttribute": 0,
        "preReadRedundancyCheckEnabled": False,
        "protectionType": "type0Protection",
        "applicationTagOwned": True,
        "maxVirtualCapacity": "69269232549888",
        "initialProvisionedCapacity": "4294967296",
        "currentProvisionedCapacity": "4294967296",
        "provisionedCapacityQuota": "55834574848",
        "growthAlertThreshold": 85,
        "expansionPolicy": "automatic",
        "volumeCache": {
            "cwob": False,
            "enterpriseCacheDump": False,
            "mirrorActive": True,
            "mirrorEnable": True,
            "readCacheActive": True,
            "readCacheEnable": True,
            "writeCacheActive": True,
            "writeCacheEnable": True,
            "cacheFlushModifier": "flush10Sec",
            "readAheadMultiplier": 0
        },
        "offline": False,
        "volumeFull": False,
        "volumeGroupRef": "0400000060080E50002998A00000945155C37C08",
        "blkSize": 512,
        "storageVolumeRef": "0200000060080E500029347000001D7855C3791D",
        "volumeCopyTarget": False,
        "volumeCopySource": False,
        "pitBaseVolume": False,
        "asyncMirrorTarget": False,
        "asyncMirrorSource": False,
        "remoteMirrorSource": False,
        "remoteMirrorTarget": False,
        "flashCached": False,
        "mediaScan": {
            "enable": False,
            "parityValidationEnable": False
        },
        "metadata": [],
        "dataAssurance": False,
        "name": "2",
        "id": "3A00000060080E500029347000001D7B55C3791E",
        "wwn": "60080E500029347000001D7B55C3791E",
        "objectType": "thinVolume",
        "mapped": False,
        "diskPool": True,
        "preferredControllerId": "070000000000000000000002",
        "totalSizeInBytes": "53687091200",
        "onlineVolumeCopy": False,
        "listOfMappings": [],
        "currentControllerId": "070000000000000000000002",
        "segmentSize": 131072,
        "cacheSettings": {
            "cwob": False,
            "enterpriseCacheDump": False,
            "mirrorActive": True,
            "mirrorEnable": True,
            "readCacheActive": True,
            "readCacheEnable": True,
            "writeCacheActive": True,
            "writeCacheEnable": True,
            "cacheFlushModifier": "flush10Sec",
            "readAheadMultiplier": 0
        },
        "thinProvisioned": True
    }
]

VOLUME = VOLUMES[0]

STORAGE_POOL = {
    'label': 'DDP',
    'id': 'fakevolgroupref',
    'volumeGroupRef': 'fakevolgroupref',
    'raidLevel': 'raidDiskPool',
    'usedSpace': '16413217521664',
    'totalRaidedSpace': '16637410312192',
}

INITIATOR_NAME = 'iqn.1998-01.com.vmware:localhost-28a58148'
INITIATOR_NAME_2 = 'iqn.1998-01.com.vmware:localhost-28a58149'
INITIATOR_NAME_3 = 'iqn.1998-01.com.vmware:localhost-28a58150'
WWPN = '20130080E5322230'
WWPN_2 = '20230080E5322230'

FC_TARGET_WWPNS = [
    '500a098280feeba5',
    '500a098290feeba5',
    '500a098190feeba5',
    '500a098180feeba5'
]

FC_I_T_MAP = {
    '20230080E5322230': [
        '500a098280feeba5',
        '500a098290feeba5'
    ],
    '20130080E5322230': [
        '500a098190feeba5',
        '500a098180feeba5'
    ]
}

FC_FABRIC_MAP = {
    'fabricB': {
        'target_port_wwn_list': [
            '500a098190feeba5',
            '500a098180feeba5'
        ],
        'initiator_port_wwn_list': [
            '20130080E5322230'
        ]
    },
    'fabricA': {
        'target_port_wwn_list': [
            '500a098290feeba5',
            '500a098280feeba5'
        ],
        'initiator_port_wwn_list': [
            '20230080E5322230'
        ]
    }
}

HOST = {
    'isSAControlled': False,
    'confirmLUNMappingCreation': False,
    'label': 'stlrx300s7-55',
    'isLargeBlockFormatHost': False,
    'clusterRef': '8500000060080E500023C7340036035F515B78FC',
    'protectionInformationCapableAccessMethod': False,
    'ports': [],
    'hostRef': '8400000060080E500023C73400300381515BFBA3',
    'hostTypeIndex': 6,
    'hostSidePorts': [{
        'label': 'NewStore',
        'type': 'iscsi',
        'address': INITIATOR_NAME}]
}
HOST_2 = {
    'isSAControlled': False,
    'confirmLUNMappingCreation': False,
    'label': 'stlrx300s7-55',
    'isLargeBlockFormatHost': False,
    'clusterRef': utils.NULL_REF,
    'protectionInformationCapableAccessMethod': False,
    'ports': [],
    'hostRef': '8400000060080E500023C73400300381515BFBA5',
    'hostTypeIndex': 6,
    'hostSidePorts': [{
        'label': 'NewStore', 'type': 'iscsi',
        'address': INITIATOR_NAME_2}]
}
# HOST_3 has all lun_ids in use.
HOST_3 = {
    'isSAControlled': False,
    'confirmLUNMappingCreation': False,
    'label': 'stlrx300s7-55',
    'isLargeBlockFormatHost': False,
    'clusterRef': '8500000060080E500023C73400360351515B78FC',
    'protectionInformationCapableAccessMethod': False,
    'ports': [],
    'hostRef': '8400000060080E501023C73400800381515BFBA5',
    'hostTypeIndex': 6,
    'hostSidePorts': [{
        'label': 'NewStore', 'type': 'iscsi',
        'address': INITIATOR_NAME_3}],
}


VOLUME_MAPPING = {
    'lunMappingRef': '8800000000000000000000000000000000000000',
    'lun': 1,
    'ssid': 16384,
    'perms': 15,
    'volumeRef': VOLUME['volumeRef'],
    'type': 'all',
    'mapRef': HOST['hostRef']
}
# VOLUME_MAPPING_3 corresponding to HOST_3 has all lun_ids in use.
VOLUME_MAPPING_3 = {
    'lunMappingRef': '8800000000000000000000000000000000000000',
    'lun': range(255),
    'ssid': 16384,
    'perms': 15,
    'volumeRef': VOLUME['volumeRef'],
    'type': 'all',
    'mapRef': HOST_3['hostRef'],
}

VOLUME_MAPPING_TO_MULTIATTACH_GROUP = copy.deepcopy(VOLUME_MAPPING)
VOLUME_MAPPING_TO_MULTIATTACH_GROUP.update(
    {'mapRef': MULTIATTACH_HOST_GROUP['clusterRef']}
)

STORAGE_SYSTEM = {
    'chassisSerialNumber': 1,
    'fwVersion': '08.10.15.00',
    'freePoolSpace': 11142431623168,
    'driveCount': 24,
    'hostSparesUsed': 0, 'id':
    '1fa6efb5-f07b-4de4-9f0e-52e5f7ff5d1b',
    'hotSpareSizeAsString': '0', 'wwn':
    '60080E500023C73400000000515AF323',
    'passwordStatus': 'valid',
    'parameters': {
        'minVolSize': 1048576, 'maxSnapshotsPerBase': 16,
        'maxDrives': 192,
        'maxVolumes': 512,
        'maxVolumesPerGroup': 256,
        'maxMirrors': 0,
        'maxMappingsPerVolume': 1,
        'maxMappableLuns': 256,
        'maxVolCopys': 511,
        'maxSnapshots': 256
    }, 'hotSpareCount': 0,
    'hostSpareCountInStandby': 0,
    'status': 'needsattn',
    'trayCount': 1,
    'usedPoolSpaceAsString': '5313000380416',
    'ip2': '10.63.165.216',
    'ip1': '10.63.165.215',
    'freePoolSpaceAsString': '11142431623168',
    'types': 'SAS',
    'name': 'stle2600-7_8',
    'hotSpareSize': 0,
    'usedPoolSpace': 5313000380416,
    'driveTypes': ['sas'],
    'unconfiguredSpaceByDriveType': {},
    'unconfiguredSpaceAsStrings': '0',
    'model': '2650',
    'unconfiguredSpace': 0
}

SNAPSHOT_GROUP = {
    'id': '3300000060080E500023C7340000098D5294AC9A',
    'status': 'optimal',
    'autoDeleteLimit': 0,
    'maxRepositoryCapacity': '-65536',
    'rollbackStatus': 'none',
    'unusableRepositoryCapacity': '0',
    'pitGroupRef': '3300000060080E500023C7340000098D5294AC9A',
    'clusterSize': 65536,
    'label': 'C6JICISVHNG2TFZX4XB5ZWL7F',
    'maxBaseCapacity': '476187142128128',
    'repositoryVolume': '3600000060080E500023BB3400001FA952CEF12C',
    'fullWarnThreshold': 99,
    'repFullPolicy': 'purgepit',
    'action': 'none',
    'rollbackPriority': 'medium',
    'creationPendingStatus': 'none',
    'consistencyGroupRef': '0000000000000000000000000000000000000000',
    'volumeHandle': 49153,
    'consistencyGroup': False,
    'baseVolume': '0200000060080E500023C734000009825294A534',
    'snapshotCount': 32
}

SNAPSHOT_IMAGE = {
    'id': fake.SNAPSHOT_ID,
    'baseVol': '0200000060080E500023C734000009825294A534',
    'status': 'optimal',
    'pitCapacity': '2147483648',
    'pitTimestamp': '1389315375',
    'pitGroupRef': '3300000060080E500023C7340000098D5294AC9A',
    'creationMethod': 'user',
    'repositoryCapacityUtilization': '2818048',
    'activeCOW': True,
    'isRollbackSource': False,
    'pitRef': '3400000060080E500023BB3400631F335294A5A8',
    'pitSequenceNumber': '19',
    'consistencyGroupId': '0000000000000000000000000000000000000000',
}

SNAPSHOT_VOLUME = {
    'id': '35000000600A0980006077F80000F8BF566581AA',
    'viewRef': '35000000600A0980006077F80000F8BF566581AA',
    'worldWideName': '600A0980006077F80000F8BF566581AA',
    'baseVol': '02000000600A0980006077F80000F89B56657E26',
    'basePIT': '0000000000000000000000000000000000000000',
    'boundToPIT': False,
    'accessMode': 'readOnly',
    'label': 'UZJ45SLUKNGWRF3QZHBTOG4C4E_DEL',
    'status': 'stopped',
    'currentManager': '070000000000000000000001',
    'preferredManager': '070000000000000000000001',
    'repositoryVolume': '0000000000000000000000000000000000000000',
    'fullWarnThreshold': 0,
    'viewTime': '1449453419',
    'viewSequenceNumber': '2104',
    'volumeHandle': 16510,
    'clusterSize': 0,
    'maxRepositoryCapacity': '0',
    'unusableRepositoryCapacity': '0',
    'membership': {
        'viewType': 'individual',
        'cgViewRef': None
    },
    'mgmtClientAttribute': 0,
    'offline': False,
    'volumeFull': False,
    'repositoryCapacity': '0',
    'baseVolumeCapacity': '1073741824',
    'totalSizeInBytes': '0',
    'consistencyGroupId': None,
    'volumeCopyTarget': False,
    'cloneCopy': False,
    'volumeCopySource': False,
    'pitBaseVolume': False,
    'asyncMirrorTarget': False,
    'asyncMirrorSource': False,
    'protectionType': 'type0Protection',
    'remoteMirrorSource': False,
    'remoteMirrorTarget': False,
    'wwn': '600A0980006077F80000F8BF566581AA',
    'listOfMappings': [],
    'mapped': False,
    'currentControllerId': '070000000000000000000001',
    'preferredControllerId': '070000000000000000000001',
    'onlineVolumeCopy': False,
    'objectType': 'pitView',
    'name': 'UZJ45SLUKNGWRF3QZHBTOG4C4E',
}

FAKE_BACKEND_STORE = {
    'key': 'cinder-snapshots',
    'value': '{"3300000060080E50003416400000E90D56B047E5":"2"}'
}

HARDWARE_INVENTORY_SINGLE_CONTROLLER = {
    'controllers': [
        {
            'modelName': '2752',
            'serialNumber': '021436001321'
        }
    ]
}

HARDWARE_INVENTORY = {
    'controllers': [
        {
            'modelName': '2752',
            'serialNumber': '021436000943'
        },
        {
            'modelName': '2752',
            'serialNumber': '021436001321'
        }
    ],
    'iscsiPorts': [
        {
            'controllerId':
            '070000000000000000000002',
            'ipv4Enabled': True,
            'ipv4Data': {
                'ipv4Address': '0.0.0.0',
                'ipv4AddressConfigMethod':
                'configStatic',
                'ipv4VlanId': {
                    'isEnabled': False,
                    'value': 0
                },
                'ipv4AddressData': {
                    'ipv4Address': '172.20.123.66',
                    'ipv4SubnetMask': '255.255.255.0',
                    'configState': 'configured',
                    'ipv4GatewayAddress': '0.0.0.0'
                }
            },
            'tcpListenPort': 3260,
            'interfaceRef': '2202040000000000000000000000000000000000',
            'iqn': 'iqn.1992-01.com.lsi:2365.60080e500023c73400000000515af323'
        }
    ],
    'fibrePorts': [
        {
            "channel": 1,
            "loopID": 126,
            "speed": 800,
            "hardAddress": 6,
            "nodeName": "20020080E5322230",
            "portName": "20130080E5322230",
            "portId": "011700",
            "topology": "fabric",
            "part": "PM8032          ",
            "revision": 8,
            "chanMiswire": False,
            "esmMiswire": False,
            "linkStatus": "up",
            "isDegraded": False,
            "speedControl": "auto",
            "maxSpeed": 800,
            "speedNegError": False,
            "reserved1": "000000000000000000000000",
            "reserved2": "",
            "ddsChannelState": 0,
            "ddsStateReason": 0,
            "ddsStateWho": 0,
            "isLocal": True,
            "channelPorts": [],
            "currentInterfaceSpeed": "speed8gig",
            "maximumInterfaceSpeed": "speed8gig",
            "interfaceRef": "2202020000000000000000000000000000000000",
            "physicalLocation": {
                "trayRef": "0000000000000000000000000000000000000000",
                "slot": 0,
                "locationParent": {
                    "refType": "generic",
                    "controllerRef": None,
                    "symbolRef": "0000000000000000000000000000000000000000",
                    "typedReference": None
                },
                "locationPosition": 0
            },
            "isTrunkCapable": False,
            "trunkMiswire": False,
            "protectionInformationCapable": True,
            "controllerId": "070000000000000000000002",
            "interfaceId": "2202020000000000000000000000000000000000",
            "addressId": "20130080E5322230",
            "niceAddressId": "20:13:00:80:E5:32:22:30"
        },
        {
            "channel": 2,
            "loopID": 126,
            "speed": 800,
            "hardAddress": 7,
            "nodeName": "20020080E5322230",
            "portName": "20230080E5322230",
            "portId": "011700",
            "topology": "fabric",
            "part": "PM8032          ",
            "revision": 8,
            "chanMiswire": False,
            "esmMiswire": False,
            "linkStatus": "up",
            "isDegraded": False,
            "speedControl": "auto",
            "maxSpeed": 800,
            "speedNegError": False,
            "reserved1": "000000000000000000000000",
            "reserved2": "",
            "ddsChannelState": 0,
            "ddsStateReason": 0,
            "ddsStateWho": 0,
            "isLocal": True,
            "channelPorts": [],
            "currentInterfaceSpeed": "speed8gig",
            "maximumInterfaceSpeed": "speed8gig",
            "interfaceRef": "2202030000000000000000000000000000000000",
            "physicalLocation": {
                "trayRef": "0000000000000000000000000000000000000000",
                "slot": 0,
                "locationParent": {
                    "refType": "generic",
                    "controllerRef": None,
                    "symbolRef": "0000000000000000000000000000000000000000",
                    "typedReference": None
                },
                "locationPosition": 0
            },
            "isTrunkCapable": False,
            "trunkMiswire": False,
            "protectionInformationCapable": True,
            "controllerId": "070000000000000000000002",
            "interfaceId": "2202030000000000000000000000000000000000",
            "addressId": "20230080E5322230",
            "niceAddressId": "20:23:00:80:E5:32:22:30"
        },
    ]
}

FAKE_POOL_ACTION_PROGRESS = [
    {
        "volumeRef": "0200000060080E50002998A00000945355C37C19",
        "progressPercentage": 55,
        "estimatedTimeToCompletion": 1,
        "currentAction": "initializing"
    },
    {
        "volumeRef": "0200000060080E50002998A00000945355C37C18",
        "progressPercentage": 0,
        "estimatedTimeToCompletion": 0,
        "currentAction": "progressDve"
    },
]

FAKE_CHAP_SECRET = 'password123'
FAKE_RESOURCE_URL = '/devmgr/v2/devmgr/utils/about'
FAKE_APP_VERSION = '2015.2|2015.2.dev59|vendor|Linux-3.13.0-24-generic'
FAKE_BACKEND = 'eseriesiSCSI'
FAKE_CINDER_HOST = 'ubuntu-1404'
FAKE_SERIAL_NUMBERS = ['021436000943', '021436001321']
FAKE_SERIAL_NUMBER = ['021436001321']
FAKE_DEFAULT_SERIAL_NUMBER = ['unknown', 'unknown']
FAKE_DEFAULT_MODEL = 'unknown'
FAKE_ABOUT_RESPONSE = {
    'runningAsProxy': True,
    'version': '01.53.9010.0005',
    'systemId': 'a89355ab-692c-4d4a-9383-e249095c3c0',
}

FAKE_TARGET_IQN = 'iqn.1992-01.com.lsi:2365.60080e500023c73400000000515af323'

FAKE_CHAP_USERNAME = 'eserieschapuser'

FAKE_CHAP_PARAMETERS = {
    'ChapAuthentication': True,
    'iqn': FAKE_TARGET_IQN,
    'chapSecret': FAKE_CHAP_SECRET,
    'authMethod': 'CHAP',
}

FAKE_CLIENT_CHAP_PARAMETERS = (
    FAKE_TARGET_IQN,
    FAKE_CHAP_USERNAME,
    FAKE_CHAP_SECRET,
)

FAKE_TARGET_DICT = {
    'data': {
        'auth_method': 'CHAP',
        'auth_password': FAKE_CHAP_SECRET,
        'auth_username': FAKE_CHAP_USERNAME,
        'discovery_auth_method': 'CHAP',
        'discovery_auth_password': FAKE_CHAP_SECRET,
        'discovery_auth_username': FAKE_CHAP_USERNAME,
        'target_discovered': False,
        'target_iqn': FAKE_TARGET_IQN,
        'target_lun': 1,
        'target_portal': '172.20.123.66:3260',
        'volume_id': '114774fb-e15a-4fae-8ee2-c9723e3645ef',
    },
    'driver_volume_type': 'iscsi',
}

FAKE_CHAP_POST_DATA = {
    'enableChapAuthentication': True,
    'alias': FAKE_CHAP_USERNAME,
    'iqn': FAKE_TARGET_IQN,
    'chapSecret': FAKE_CHAP_SECRET,
    'authMethod': 'CHAP',
}


FAKE_CONTROLLERS = [
    {'serialNumber': FAKE_SERIAL_NUMBERS[0], 'modelName': '2752'},
    {'serialNumber': FAKE_SERIAL_NUMBERS[1], 'modelName': '2752'}]

FAKE_SINGLE_CONTROLLER = [{'serialNumber': FAKE_SERIAL_NUMBERS[1]}]

FAKE_KEY = ('openstack-%s-%s-%s' % (FAKE_CINDER_HOST, FAKE_SERIAL_NUMBERS[0],
                                    FAKE_SERIAL_NUMBERS[1]))

FAKE_ASUP_DATA = {
    'category': 'provisioning',
    'app-version': FAKE_APP_VERSION,
    'event-source': 'Cinder driver NetApp_iSCSI_ESeries',
    'event-description': 'OpenStack Cinder connected to E-Series proxy',
    'system-version': '08.10.15.00',
    'computer-name': FAKE_CINDER_HOST,
    'model': FAKE_CONTROLLERS[0]['modelName'],
    'controller2-serial': FAKE_CONTROLLERS[1]['serialNumber'],
    'controller1-serial': FAKE_CONTROLLERS[0]['serialNumber'],
    'chassis-serial-number': FAKE_SERIAL_NUMBER[0],
    'operating-mode': 'proxy',
}

GET_ASUP_RETURN = {
    'model': FAKE_CONTROLLERS[0]['modelName'],
    'serial_numbers': FAKE_SERIAL_NUMBERS,
    'firmware_version': FAKE_ASUP_DATA['system-version'],
    'chassis_sn': FAKE_ASUP_DATA['chassis-serial-number'],
}

FAKE_POST_INVOKE_DATA = ('POST', '/key-values/%s' % FAKE_KEY,
                         json.dumps(FAKE_ASUP_DATA))

VOLUME_COPY_JOB = {
    "status": "complete",
    "cloneCopy": True,
    "pgRef": "3300000060080E500023C73400000ACA52D29454",
    "volcopyHandle": 49160,
    "idleTargetWriteProt": True,
    "copyPriority": "priority2",
    "volcopyRef": "1800000060080E500023C73400000ACF52D29466",
    "worldWideName": "60080E500023C73400000ACF52D29466",
    "copyCompleteTime": "0",
    "sourceVolume": "3500000060080E500023C73400000ACE52D29462",
    "currentManager": "070000000000000000000002",
    "copyStartTime": "1389551671",
    "reserved1": "00000000",
    "targetVolume": "0200000060080E500023C73400000A8C52D10675",
}

FAKE_ENDPOINT_HTTP = 'http://host:80/endpoint'

FAKE_ENDPOINT_HTTPS = 'https://host:8443/endpoint'

FAKE_INVOC_MSG = 'success'

FAKE_CLIENT_PARAMS = {
    'scheme': 'http',
    'host': '127.0.0.1',
    'port': 8080,
    'service_path': '/devmgr/vn',
    'username': 'rw',
    'password': 'rw',
}

FAKE_CONSISTENCY_GROUP = {
    'cgRef': '2A000000600A0980006077F8008702F45480F41A',
    'label': '5BO5GPO4PFGRPMQWEXGTILSAUI',
    'repFullPolicy': 'failbasewrites',
    'fullWarnThreshold': 75,
    'autoDeleteLimit': 0,
    'rollbackPriority': 'medium',
    'uniqueSequenceNumber': [8940, 8941, 8942],
    'creationPendingStatus': 'none',
    'name': '5BO5GPO4PFGRPMQWEXGTILSAUI',
    'id': '2A000000600A0980006077F8008702F45480F41A'
}

FAKE_CONSISTENCY_GROUP_MEMBER = {
    'consistencyGroupId': '2A000000600A0980006077F8008702F45480F41A',
    'volumeId': '02000000600A0980006077F8000002F55480F421',
    'volumeWwn': '600A0980006077F8000002F55480F421',
    'baseVolumeName': 'I5BHHNILUJGZHEUD4S36GCOQYA',
    'clusterSize': 65536,
    'totalRepositoryVolumes': 1,
    'totalRepositoryCapacity': '4294967296',
    'usedRepositoryCapacity': '5636096',
    'fullWarnThreshold': 75,
    'totalSnapshotImages': 3,
    'totalSnapshotVolumes': 2,
    'autoDeleteSnapshots': False,
    'autoDeleteLimit': 0,
    'pitGroupId': '33000000600A0980006077F8000002F85480F435',
    'repositoryVolume': '36000000600A0980006077F8000002F75480F435'
}
FAKE_CONSISTENCY_GROUP_SNAPSHOT_VOLUME = {
    'id': '2C00000060080E500034194F002C96A256BD50F9',
    'name': '6TRZHKDG75DVLBC2JU5J647RME',
    'cgViewRef': '2C00000060080E500034194F002C96A256BD50F9',
    'groupRef': '2A00000060080E500034194F0087969856BD2D67',
    'label': '6TRZHKDG75DVLBC2JU5J647RME',
    'viewTime': '1455221060',
    'viewSequenceNumber': '10',
}


def list_snapshot_groups(numGroups):
    snapshots = []
    for n in range(0, numGroups):
        s = copy.deepcopy(SNAPSHOT_GROUP)
        s['label'] = s['label'][:-1] + str(n)
        snapshots.append(s)
    return snapshots


def create_configuration_eseries():
    config = conf.Configuration(None)
    config.append_config_values(na_opts.netapp_connection_opts)
    config.append_config_values(na_opts.netapp_transport_opts)
    config.append_config_values(na_opts.netapp_basicauth_opts)
    config.append_config_values(na_opts.netapp_provisioning_opts)
    config.append_config_values(na_opts.netapp_eseries_opts)
    config.netapp_storage_protocol = 'iscsi'
    config.netapp_login = 'rw'
    config.netapp_password = 'rw'
    config.netapp_server_hostname = '127.0.0.1'
    config.netapp_transport_type = 'http'
    config.netapp_server_port = '8080'
    config.netapp_storage_pools = 'DDP'
    config.netapp_storage_family = 'eseries'
    config.netapp_sa_password = 'saPass'
    config.netapp_controller_ips = '10.11.12.13,10.11.12.14'
    config.netapp_webservice_path = '/devmgr/v2'
    config.netapp_enable_multiattach = False
    config.use_chap_auth = False
    return config


def deepcopy_return_value_method_decorator(fn):
    """Returns a deepcopy of the returned value of the wrapped function."""
    def decorator(*args, **kwargs):
        return copy.deepcopy(fn(*args, **kwargs))

    return decorator


def deepcopy_return_value_class_decorator(cls):
    """Wraps 'non-protected' methods of a class with decorator.

    Wraps all 'non-protected' methods of a class with the
    deepcopy_return_value_method_decorator decorator.
    """
    class NewClass(cls):
        def __getattribute__(self, attr_name):
            obj = super(NewClass, self).__getattribute__(attr_name)
            if (hasattr(obj, '__call__') and not attr_name.startswith('_')
                    and not isinstance(obj, mock.Mock)):
                return deepcopy_return_value_method_decorator(obj)
            return obj

    return NewClass


@deepcopy_return_value_class_decorator
class FakeEseriesClient(object):
    features = na_utils.Features()

    def __init__(self, *args, **kwargs):
        self.features.add_feature('AUTOSUPPORT')
        self.features.add_feature('SSC_API_V2')
        self.features.add_feature('REST_1_3_RELEASE')
        self.features.add_feature('REST_1_4_RELEASE')

    def list_storage_pools(self):
        return STORAGE_POOLS

    def register_storage_system(self, *args, **kwargs):
        return {
            'freePoolSpace': '17055871480319',
            'driveCount': 24,
            'wwn': '60080E500023C73400000000515AF323',
            'id': '1',
            'hotSpareSizeAsString': '0',
            'hostSparesUsed': 0,
            'types': '',
            'hostSpareCountInStandby': 0,
            'status': 'optimal',
            'trayCount': 1,
            'usedPoolSpaceAsString': '37452115456',
            'ip2': '10.63.165.216',
            'ip1': '10.63.165.215',
            'freePoolSpaceAsString': '17055871480319',
            'hotSpareCount': 0,
            'hotSpareSize': '0',
            'name': 'stle2600-7_8',
            'usedPoolSpace': '37452115456',
            'driveTypes': ['sas'],
            'unconfiguredSpaceByDriveType': {},
            'unconfiguredSpaceAsStrings': '0',
            'model': '2650',
            'unconfiguredSpace': '0'
        }

    def list_volume(self, volume_id):
        return VOLUME

    def list_volumes(self):
        return [VOLUME]

    def delete_volume(self, vol):
        pass

    def create_host_group(self, name):
        return MULTIATTACH_HOST_GROUP

    def get_host_group(self, ref):
        return MULTIATTACH_HOST_GROUP

    def list_host_groups(self):
        return [MULTIATTACH_HOST_GROUP, FOREIGN_HOST_GROUP]

    def get_host_group_by_name(self, name, *args, **kwargs):
        host_groups = self.list_host_groups()
        return [host_group for host_group in host_groups
                if host_group['label'] == name][0]

    def set_host_group_for_host(self, *args, **kwargs):
        pass

    def create_host_with_ports(self, *args, **kwargs):
        return HOST

    def list_hosts(self):
        return [HOST, HOST_2]

    def get_host(self, *args, **kwargs):
        return HOST

    def create_volume(self, *args, **kwargs):
        return VOLUME

    def create_volume_mapping(self, *args, **kwargs):
        return VOLUME_MAPPING

    def get_volume_mappings(self):
        return [VOLUME_MAPPING]

    def get_volume_mappings_for_volume(self, volume):
        return [VOLUME_MAPPING]

    def get_volume_mappings_for_host(self, host_ref):
        return [VOLUME_MAPPING]

    def get_volume_mappings_for_host_group(self, hg_ref):
        return [VOLUME_MAPPING]

    def delete_volume_mapping(self):
        return

    def move_volume_mapping_via_symbol(self, map_ref, to_ref, lun_id):
        return {'lun': lun_id}

    def list_storage_system(self):
        return STORAGE_SYSTEM

    def list_storage_systems(self):
        return [STORAGE_SYSTEM]

    def list_snapshot_groups(self):
        return [SNAPSHOT_GROUP]

    def list_snapshot_images(self):
        return [SNAPSHOT_IMAGE]

    def list_snapshot_image(self, *args, **kwargs):
        return SNAPSHOT_IMAGE

    def create_cg_snapshot_view(self, *args, **kwargs):
        return SNAPSHOT_VOLUME

    def list_host_types(self):
        return [
            {
                'name': 'FactoryDefault',
                'index': 0,
                'code': 'FactoryDefault',
            },
            {
                'name': 'Windows 2000/Server 2003/Server 2008 Non-Clustered',
                'index': 1,
                'code': 'W2KNETNCL',
            },
            {
                'name': 'Solaris',
                'index': 2,
                'code': 'SOL',
            },
            {
                'name': 'ONTAP_RDAC',
                'index': 4,
                'code': 'ONTAP_RDAC',
            },
            {
                'name': 'AVT_4M',
                'index': 5,
                'code': 'AVT_4M',
            },
            {
                'name': 'Linux',
                'index': 6,
                'code': 'LNX',
            },
            {
                'name': 'LnxALUA',
                'index': 7,
                'code': 'LnxALUA',
            },
            {
                'name': 'Windows 2000/Server 2003/Server 2008 Clustered',
                'index': 8,
                'code': 'W2KNETCL',
            },
            {
                'name': 'AIX MPIO',
                'index': 9,
                'code': 'AIX MPIO',
            },
            {
                'name': 'VmwTPGSALUA',
                'index': 10,
                'code': 'VmwTPGSALUA',
            },
            {
                'name': 'HP-UX TPGS',
                'index': 15,
                'code': 'HPXTPGS',
            },
            {
                'name': 'SolTPGSALUA',
                'index': 17,
                'code': 'SolTPGSALUA',
            },
            {
                'name': 'SVC',
                'index': 18,
                'code': 'SVC',
            },
            {
                'name': 'MacTPGSALUA',
                'index': 22,
                'code': 'MacTPGSALUA',
            },
            {
                'name': 'WinTPGSALUA',
                'index': 23,
                'code': 'WinTPGSALUA',
            },
            {
                'name': 'LnxTPGSALUA',
                'index': 24,
                'code': 'LnxTPGSALUA',
            },
            {
                'name': 'LnxTPGSALUA_PM',
                'index': 25,
                'code': 'LnxTPGSALUA_PM',
            },
            {
                'name': 'ONTAP_ALUA',
                'index': 26,
                'code': 'ONTAP_ALUA',
            },
            {
                'name': 'LnxTPGSALUA_SF',
                'index': 27,
                'code': 'LnxTPGSALUA_SF',
            }
        ]

    def update_host_type(self, *args, **kwargs):
        pass

    def list_hardware_inventory(self):
        return HARDWARE_INVENTORY

    def get_eseries_api_info(self, verify=False):
        return 'Proxy', '1.53.9010.0005'

    def set_counter(self, key, value):
        pass

    def add_autosupport_data(self, *args):
        pass

    def set_chap_authentication(self, *args, **kwargs):
        return FAKE_CHAP_PARAMETERS

    def get_serial_numbers(self):
        return FAKE_ASUP_DATA.get('controller1-serial'), FAKE_ASUP_DATA.get(
            'controller2-serial')

    def get_model_name(self):
        pass

    def api_operating_mode(self):
        pass

    def get_firmware_version(self):
        return FAKE_ASUP_DATA['system-version']

    def create_volume_copy_job(self, *args, **kwargs):
        return VOLUME_COPY_JOB

    def list_vol_copy_job(self, *args, **kwargs):
        return VOLUME_COPY_JOB

    def delete_vol_copy_job(self, *args, **kwargs):
        pass

    def create_snapshot_image(self, *args, **kwargs):
        return SNAPSHOT_IMAGE

    def create_snapshot_volume(self, *args, **kwargs):
        return SNAPSHOT_VOLUME

    def list_snapshot_volumes(self, *args, **kwargs):
        return [SNAPSHOT_VOLUME]

    def list_snapshot_volume(self, *args, **kwargs):
        return SNAPSHOT_IMAGE

    def create_snapshot_group(self, *args, **kwargs):
        return SNAPSHOT_GROUP

    def list_snapshot_group(self, *args, **kwargs):
        return SNAPSHOT_GROUP

    def delete_snapshot_volume(self, *args, **kwargs):
        pass

    def list_target_wwpns(self, *args, **kwargs):
        return [WWPN_2]

    def update_stored_system_password(self, *args, **kwargs):
        pass

    def update_snapshot_volume(self, *args, **kwargs):
        return SNAPSHOT_VOLUME

    def delete_snapshot_image(self, *args, **kwargs):
        pass

    def delete_snapshot_group(self, *args, **kwargs):
        pass

    def restart_snapshot_volume(self, *args, **kwargs):
        pass

    def create_consistency_group(self, *args, **kwargs):
        return FAKE_CONSISTENCY_GROUP

    def delete_consistency_group(self, *args, **kwargs):
        pass

    def list_consistency_groups(self, *args, **kwargs):
        return [FAKE_CONSISTENCY_GROUP]

    def remove_consistency_group_member(self, *args, **kwargs):
        pass

    def add_consistency_group_member(self, *args, **kwargs):
        pass

    def list_backend_store(self, key):
        return {}

    def save_backend_store(self, key, val):
        pass

    def create_consistency_group_snapshot(self, *args, **kwargs):
        return [SNAPSHOT_IMAGE]

    def get_consistency_group_snapshots(self, *args, **kwargs):
        return [SNAPSHOT_IMAGE]

    def delete_consistency_group_snapshot(self, *args, **kwargs):
        pass



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common


class VirtualArray(common.CoprHDResource):

    # Commonly used URIs for the 'varrays' module
    URI_VIRTUALARRAY = '/vdc/varrays'
    URI_VIRTUALARRAY_BY_VDC_ID = '/vdc/varrays?vdc-id={0}'
    URI_VIRTUALARRAY_URI = '/vdc/varrays/{0}'

    def varray_query(self, name):
        """Returns the UID of the varray specified by the name."""
        if common.is_uri(name):
            return name

        uris = self.varray_list()

        for uri in uris:
            varray = self.varray_show(uri)
            if varray and varray['name'] == name:
                return varray['id']

        raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                 (_("varray %s: not found") % name))

    def varray_list(self, vdcname=None):
        """Returns all the varrays in a vdc.

        :param vdcname: Name of the Virtual Data Center
        :returns: JSON payload of varray list
        """
        vdcrestapi = None
        if vdcname is not None:
            vdcrestapi = VirtualArray.URI_VIRTUALARRAY_BY_VDC_ID.format(
                vdcname)
        else:
            vdcrestapi = VirtualArray.URI_VIRTUALARRAY
        (s, h) = common.service_json_request(
            self.ipaddr, self.port, "GET",
            vdcrestapi, None)

        o = common.json_decode(s)

        returnlst = []
        for item in o['varray']:
            returnlst.append(item['id'])

        return returnlst

    def varray_show(self, label):
        """Makes REST API call to retrieve varray details based on name."""
        uri = self.varray_query(label)

        (s, h) = common.service_json_request(
            self.ipaddr, self.port, "GET",
            VirtualArray.URI_VIRTUALARRAY_URI.format(uri),
            None)

        o = common.json_decode(s)
        if 'inactive' in o and o['inactive'] is True:
            return None
        else:
            return o



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_utils import timeutils
import six
from sqlalchemy import MetaData, Table

from cinder.volume import group_types as volume_group_types


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    now = timeutils.utcnow()

    group_types = Table('group_types', meta, autoload=True)
    group_type_specs = Table('group_type_specs', meta, autoload=True)

    # Create a default group_type for migrating cgsnapshots
    results = list(group_types.select().where(
        group_types.c.name == volume_group_types.DEFAULT_CGSNAPSHOT_TYPE and
        group_types.c.deleted is False).
        execute())
    if not results:
        grp_type_id = six.text_type(uuid.uuid4())
        group_type_dicts = {
            'id': grp_type_id,
            'name': volume_group_types.DEFAULT_CGSNAPSHOT_TYPE,
            'description': 'Default group type for migrating cgsnapshot',
            'created_at': now,
            'updated_at': now,
            'deleted': False,
            'is_public': True,
        }
        grp_type = group_types.insert()
        grp_type.execute(group_type_dicts)
    else:
        grp_type_id = results[0]['id']

    results = list(group_type_specs.select().where(
        group_type_specs.c.group_type_id == grp_type_id and
        group_type_specs.c.deleted is False).
        execute())
    if not results:
        group_spec_dicts = {
            'key': 'consistent_group_snapshot_enabled',
            'value': '<is> True',
            'group_type_id': grp_type_id,
            'created_at': now,
            'updated_at': now,
            'deleted': False,
        }
        grp_spec = group_type_specs.insert()
        grp_spec.execute(group_spec_dicts)



# Copyright (c) 2017 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fibre Channel Driver for DataCore SANsymphony storage array."""

from oslo_log import log as logging

from cinder import exception as cinder_exception
from cinder.i18n import _
from cinder import interface
from cinder import utils as cinder_utils
from cinder.volume.drivers.datacore import driver
from cinder.volume.drivers.datacore import exception as datacore_exception


LOG = logging.getLogger(__name__)


@interface.volumedriver
class FibreChannelVolumeDriver(driver.DataCoreVolumeDriver):
    """DataCore SANsymphony Fibre Channel volume driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver

    """

    VERSION = '1.0.0'
    STORAGE_PROTOCOL = 'FC'
    CI_WIKI_NAME = 'DataCore_CI'

    def __init__(self, *args, **kwargs):
        super(FibreChannelVolumeDriver, self).__init__(*args, **kwargs)

    def validate_connector(self, connector):
        """Fail if connector doesn't contain all the data needed by the driver.

        :param connector: Connector information
        """

        required_data = ['host', 'wwpns']
        for required in required_data:
            if required not in connector:
                LOG.error("The volume driver requires %(data)s "
                          "in the connector.", {'data': required})
                raise cinder_exception.InvalidConnectorException(
                    missing=required)

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info.

        :param volume: Volume object
        :param connector: Connector information
        :return: Connection information
        """

        LOG.debug("Initialize connection for volume %(volume)s for "
                  "connector %(connector)s.",
                  {'volume': volume['id'], 'connector': connector})

        virtual_disk = self._get_virtual_disk_for(volume, raise_not_found=True)

        if virtual_disk.DiskStatus != 'Online':
            LOG.warning("Attempting to attach virtual disk %(disk)s "
                        "that is in %(state)s state.",
                        {'disk': virtual_disk.Id,
                         'state': virtual_disk.DiskStatus})

        serve_result = self._serve_virtual_disk(connector, virtual_disk.Id)

        online_servers = [server.Id for server in self._get_online_servers()]
        online_ports = self._get_online_ports(online_servers)
        online_devices = self._get_online_devices(online_ports)
        online_units = [unit for unit in serve_result[1]
                        if unit.VirtualTargetDeviceId in online_devices]

        if not online_units:
            msg = (_("Volume %(volume)s can not be attached "
                     "to connector %(connector)s due to backend state.")
                   % {'volume': volume['id'], 'connector': connector})
            LOG.error(msg)
            try:
                self._api.unserve_virtual_disks_from_host(serve_result[0].Id,
                                                          [virtual_disk.Id])
            except datacore_exception.DataCoreException as e:
                LOG.warning("An error occurred on a cleanup after failed "
                            "attaching of volume %(volume)s to connector "
                            "%(connector)s: %(error)s.",
                            {'volume': volume['id'],
                             'connector': connector,
                             'error': e})
            raise cinder_exception.VolumeDriverException(message=msg)

        target_device = online_devices[online_units[0].VirtualTargetDeviceId]
        target_port = online_ports[target_device.TargetPortId]

        connection_data = {
            'target_discovered': False,
            'target_lun': online_units[0].Lun.Quad,
            'target_wwn': target_port.PortName.replace('-', '').lower(),
            'volume_id': volume['id'],
            'access_mode': 'rw',
        }

        LOG.debug("Connection data: %s", connection_data)

        return {
            'driver_volume_type': 'fibre_channel',
            'data': connection_data,
        }

    def _serve_virtual_disk(self, connector, virtual_disk_id):
        server_group = self._get_our_server_group()

        @cinder_utils.synchronized(
            'datacore-backend-%s' % server_group.Id, external=True)
        def serve_virtual_disk():
            connector_wwpns = list(wwpn.replace('-', '').lower()
                                   for wwpn in connector['wwpns'])

            client = self._get_client(connector['host'], create_new=True)

            available_ports = self._api.get_ports()

            initiators = []
            for port in available_ports:
                port_name = port.PortName.replace('-', '').lower()
                if (port.PortType == 'FibreChannel'
                        and port.PortMode == 'Initiator'
                        and port_name in connector_wwpns):
                    initiators.append(port)
            if not initiators:
                msg = _("Fibre Channel ports not found for "
                        "connector: %s") % connector
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)
            else:
                for initiator in initiators:
                    if initiator.HostId != client.Id:
                        try:
                            self._api.assign_port(client.Id, initiator.Id)
                        except datacore_exception.DataCoreException as e:
                            LOG.info("Assigning initiator port %(initiator)s "
                                     "to client %(client)s failed with "
                                     "error: %(error)s",
                                     {'initiator': initiator.Id,
                                      'client': client.Id,
                                      'error': e})

            virtual_logical_units = self._api.serve_virtual_disks_to_host(
                client.Id, [virtual_disk_id])

            return client, virtual_logical_units

        return serve_virtual_disk()

    def _get_online_ports(self, online_servers):
        ports = self._api.get_ports()
        online_ports = {port.Id: port for port in ports
                        if port.HostId in online_servers}

        return online_ports

    def _get_online_devices(self, online_ports):
        devices = self._api.get_target_devices()
        online_devices = {device.Id: device for device in devices
                          if device.TargetPortId in online_ports}

        return online_devices



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Text, Column, MetaData, Table


def upgrade(migrate_engine):
    """Add the connector column to the volume_attachment table."""
    meta = MetaData(bind=migrate_engine)
    volume_attachment = Table('volume_attachment', meta, autoload=True)
    if not hasattr(volume_attachment.c, 'connector'):
        volume_attachment.create_column(Column('connector', Text))



# Copyright (c) 2011 Zadara Storage Inc.
# Copyright (c) 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume types extra specs extension"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_log import versionutils
from six.moves import http_client
import webob

from cinder.api import common
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder import context as ctxt
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder.policies import type_extra_specs as policy
from cinder import rpc
from cinder import utils
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)

extraspec_opts = [
    cfg.BoolOpt('allow_inuse_volume_type_modification',
                default=False,
                deprecated_for_removal=True,
                help="DEPRECATED: Allow the ability to modify the "
                     "extra-spec settings of an in-use volume-type."),

]

CONF = cfg.CONF
CONF.register_opts(extraspec_opts)


class VolumeTypeExtraSpecsController(wsgi.Controller):
    """The volume type extra specs API controller for the OpenStack API."""

    def _get_extra_specs(self, context, type_id):
        extra_specs = db.volume_type_extra_specs_get(context, type_id)
        specs_dict = {}
        for key, value in extra_specs.items():
            specs_dict[key] = value
        return dict(extra_specs=specs_dict)

    def _check_type(self, context, type_id):
        # Not found exception will be handled at the wsgi level
        volume_types.get_volume_type(context, type_id)

    def index(self, req, type_id):
        """Returns the list of extra specs for a given volume type."""
        context = req.environ['cinder.context']
        context.authorize(policy.GET_ALL_POLICY)
        self._check_type(context, type_id)
        return self._get_extra_specs(context, type_id)

    def _allow_update(self, context, type_id):
        if (not CONF.allow_inuse_volume_type_modification):
            vols = db.volume_get_all(
                ctxt.get_admin_context(),
                limit=1,
                filters={'volume_type_id': type_id})
            if len(vols):
                expl = _('Volume Type is currently in use.')
                raise webob.exc.HTTPBadRequest(explanation=expl)
        else:
            msg = ("The option 'allow_inuse_volume_type_modification' "
                   "is deprecated and will be removed in a future "
                   "release.  The default behavior going forward will "
                   "be to disallow modificaton of in-use types.")
            versionutils.report_deprecated_feature(LOG, msg)
        return

    def create(self, req, type_id, body=None):
        context = req.environ['cinder.context']
        context.authorize(policy.CREATE_POLICY)
        self._allow_update(context, type_id)

        self.assert_valid_body(body, 'extra_specs')

        self._check_type(context, type_id)
        specs = body['extra_specs']
        self._check_key_names(specs.keys())
        utils.validate_dictionary_string_length(specs)

        db.volume_type_extra_specs_update_or_create(context,
                                                    type_id,
                                                    specs)
        # Get created_at and updated_at for notification
        volume_type = volume_types.get_volume_type(context, type_id)
        notifier_info = dict(type_id=type_id, specs=specs,
                             created_at=volume_type['created_at'],
                             updated_at=volume_type['updated_at'])
        notifier = rpc.get_notifier('volumeTypeExtraSpecs')
        notifier.info(context, 'volume_type_extra_specs.create',
                      notifier_info)
        return body

    def update(self, req, type_id, id, body=None):
        context = req.environ['cinder.context']
        context.authorize(policy.UPDATE_POLICY)
        self._allow_update(context, type_id)

        if not body:
            expl = _('Request body empty')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        self._check_type(context, type_id)
        if id not in body:
            expl = _('Request body and URI mismatch')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        if len(body) > 1:
            expl = _('Request body contains too many items')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        self._check_key_names(body.keys())
        utils.validate_dictionary_string_length(body)

        db.volume_type_extra_specs_update_or_create(context,
                                                    type_id,
                                                    body)
        # Get created_at and updated_at for notification
        volume_type = volume_types.get_volume_type(context, type_id)
        notifier_info = dict(type_id=type_id, id=id,
                             created_at=volume_type['created_at'],
                             updated_at=volume_type['updated_at'])
        notifier = rpc.get_notifier('volumeTypeExtraSpecs')
        notifier.info(context,
                      'volume_type_extra_specs.update',
                      notifier_info)
        return body

    def show(self, req, type_id, id):
        """Return a single extra spec item."""
        context = req.environ['cinder.context']
        context.authorize(policy.GET_POLICY)
        self._check_type(context, type_id)
        specs = self._get_extra_specs(context, type_id)
        if id in specs['extra_specs']:
            return {id: specs['extra_specs'][id]}
        else:
            raise exception.VolumeTypeExtraSpecsNotFound(
                volume_type_id=type_id, extra_specs_key=id)

    @wsgi.response(http_client.ACCEPTED)
    def delete(self, req, type_id, id):
        """Deletes an existing extra spec."""
        context = req.environ['cinder.context']
        self._check_type(context, type_id)
        context.authorize(policy.DELETE_POLICY)
        self._allow_update(context, type_id)

        # Not found exception will be handled at the wsgi level
        db.volume_type_extra_specs_delete(context, type_id, id)

        # Get created_at and updated_at for notification
        volume_type = volume_types.get_volume_type(context, type_id)
        notifier_info = dict(type_id=type_id, id=id,
                             created_at=volume_type['created_at'],
                             updated_at=volume_type['updated_at'],
                             deleted_at=volume_type['deleted_at'])
        notifier = rpc.get_notifier('volumeTypeExtraSpecs')
        notifier.info(context,
                      'volume_type_extra_specs.delete',
                      notifier_info)

    def _check_key_names(self, keys):
        if not common.validate_key_names(keys):
            expl = _('Key names can only contain alphanumeric characters, '
                     'underscores, periods, colons and hyphens.')

            raise webob.exc.HTTPBadRequest(explanation=expl)


class Types_extra_specs(extensions.ExtensionDescriptor):
    """Type extra specs support."""

    name = "TypesExtraSpecs"
    alias = "os-types-extra-specs"
    updated = "2011-08-24T00:00:00+00:00"

    def get_resources(self):
        resources = []
        res = extensions.ResourceExtension('extra_specs',
                                           VolumeTypeExtraSpecsController(),
                                           parent=dict(member_name='type',
                                                       collection_name='types')
                                           )
        resources.append(res)

        return resources



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import gettext
import os

gettext.install('cinder')

from cinder import utils


def setup(app):
    print("**Autodocumenting from %s" % os.path.abspath(os.curdir))
    rv = utils.execute('./doc/generate_autodoc_index.sh')
    print(rv[0])



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit tests for NexentaStor 5 REST API helper
"""

import uuid

import mock
from mock import patch
from oslo_serialization import jsonutils
import requests
from requests import adapters
from six.moves import http_client

from cinder import exception
from cinder import test
from cinder.volume.drivers.nexenta.ns5 import jsonrpc

HOST = '1.1.1.1'
USERNAME = 'user'
PASSWORD = 'pass'


def gen_response(code=http_client.OK, json=None):
    r = requests.Response()
    r.headers['Content-Type'] = 'application/json'
    r.encoding = 'utf8'
    r.status_code = code
    r.reason = 'FAKE REASON'
    r.raw = mock.Mock()
    r._content = ''
    if json:
        r._content = jsonutils.dumps(json)
    return r


class TestNexentaJSONProxyAuth(test.TestCase):

    @patch('cinder.volume.drivers.nexenta.ns5.jsonrpc.requests.post')
    def test_https_auth(self, post):
        use_https = True
        port = 8443
        auth_uri = 'auth/login'
        rnd_url = 'some/random/url'

        class PostSideEffect(object):
            def __call__(self, *args, **kwargs):
                r = gen_response()
                if args[0] == '%(scheme)s://%(host)s:%(port)s/%(uri)s' % {
                        'scheme': 'https',
                        'host': HOST,
                        'port': port,
                        'uri': auth_uri}:
                    token = uuid.uuid4().hex
                    content = {'token': token}
                    r._content = jsonutils.dumps(content)
                return r
        post_side_effect = PostSideEffect()
        post.side_effect = post_side_effect

        class TestAdapter(adapters.HTTPAdapter):

            def __init__(self):
                super(TestAdapter, self).__init__()
                self.counter = 0

            def send(self, request, *args, **kwargs):
                # an url is being requested for the second time
                if self.counter == 1:
                    # make the fake backend respond 401
                    r = gen_response(http_client.UNAUTHORIZED)
                    r._content = ''
                    r.connection = mock.Mock()
                    r_ = gen_response(json={'data': []})
                    r.connection.send = lambda prep, **kwargs_: r_
                else:
                    r = gen_response(json={'data': []})
                r.request = request
                self.counter += 1
                return r

        nef = jsonrpc.NexentaJSONProxy(HOST, port, USERNAME, PASSWORD,
                                       use_https)
        adapter = TestAdapter()
        nef.session.mount(
            '%(scheme)s://%(host)s:%(port)s/%(uri)s' % {
                'scheme': 'https',
                'host': HOST,
                'port': port,
                'uri': rnd_url},
            adapter)

        # successful authorization
        self.assertEqual({'data': []}, nef.get(rnd_url))

        # session timeout simulation. Client must authenticate newly
        self.assertEqual({'data': []}, nef.get(rnd_url))
        # auth URL must be requested two times at this moment
        self.assertEqual(2, post.call_count)

        # continue with the last (second) token
        self.assertEqual(nef.get(rnd_url), {'data': []})
        # auth URL must be requested two times
        self.assertEqual(2, post.call_count)


class TestNexentaJSONProxy(test.TestCase):

    def setUp(self):
        super(TestNexentaJSONProxy, self).setUp()
        self.nef = jsonrpc.NexentaJSONProxy(HOST, 0, USERNAME, PASSWORD, False)

    def gen_adapter(self, code, json=None):
        class TestAdapter(adapters.HTTPAdapter):

            def __init__(self):
                super(TestAdapter, self).__init__()

            def send(self, request, *args, **kwargs):
                r = gen_response(code, json)
                r.request = request
                return r

        return TestAdapter()

    def _mount_adapter(self, url, adapter):
        self.nef.session.mount(
            '%(scheme)s://%(host)s:%(port)s/%(uri)s' % {
                'scheme': 'http',
                'host': HOST,
                'port': 8080,
                'uri': url},
            adapter)

    def test_post(self):
        random_dict = {'data': uuid.uuid4().hex}
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, self.gen_adapter(http_client.CREATED,
                                                      random_dict))
        self.assertEqual(random_dict, self.nef.post(rnd_url))

    def test_delete(self):
        random_dict = {'data': uuid.uuid4().hex}
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, self.gen_adapter(http_client.CREATED,
                                                      random_dict))
        self.assertEqual(random_dict, self.nef.delete(rnd_url))

    def test_put(self):
        random_dict = {'data': uuid.uuid4().hex}
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, self.gen_adapter(http_client.CREATED,
                                                      random_dict))
        self.assertEqual(random_dict, self.nef.put(rnd_url))

    def test_get_200(self):
        random_dict = {'data': uuid.uuid4().hex}
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, self.gen_adapter(http_client.OK,
                                                      random_dict))
        self.assertEqual(random_dict, self.nef.get(rnd_url))

    def test_get_201(self):
        random_dict = {'data': uuid.uuid4().hex}
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, self.gen_adapter(http_client.CREATED,
                                                      random_dict))
        self.assertEqual(random_dict, self.nef.get(rnd_url))

    def test_get_500(self):
        class TestAdapter(adapters.HTTPAdapter):

            def __init__(self):
                super(TestAdapter, self).__init__()

            def send(self, request, *args, **kwargs):
                json = {
                    'code': 'NEF_ERROR',
                    'message': 'Some error'
                }
                r = gen_response(http_client.INTERNAL_SERVER_ERROR, json)
                r.request = request
                return r

        adapter = TestAdapter()
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, adapter)
        self.assertRaises(exception.NexentaException, self.nef.get, rnd_url)

    def test_get__not_nef_error(self):
        class TestAdapter(adapters.HTTPAdapter):

            def __init__(self):
                super(TestAdapter, self).__init__()

            def send(self, request, *args, **kwargs):
                r = gen_response(http_client.NOT_FOUND)
                r._content = 'Page Not Found'
                r.request = request
                return r

        adapter = TestAdapter()
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, adapter)
        self.assertRaises(exception.VolumeBackendAPIException, self.nef.get,
                          rnd_url)

    def test_get__not_nef_error_empty_body(self):
        class TestAdapter(adapters.HTTPAdapter):

            def __init__(self):
                super(TestAdapter, self).__init__()

            def send(self, request, *args, **kwargs):
                r = gen_response(http_client.NOT_FOUND)
                r.request = request
                return r

        adapter = TestAdapter()
        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, adapter)
        self.assertRaises(exception.VolumeBackendAPIException, self.nef.get,
                          rnd_url)

    def test_202(self):
        redirect_url = 'redirect/url'

        class RedirectTestAdapter(adapters.HTTPAdapter):

            def __init__(self):
                super(RedirectTestAdapter, self).__init__()

            def send(self, request, *args, **kwargs):
                json = {
                    'links': [{'href': redirect_url}]
                }
                r = gen_response(http_client.ACCEPTED, json)
                r.request = request
                return r

        rnd_url = 'some/random/url'
        self._mount_adapter(rnd_url, RedirectTestAdapter())
        self._mount_adapter(redirect_url, self.gen_adapter(
            http_client.CREATED))
        self.assertIsNone(self.nef.get(rnd_url))



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Contains tagging related methods."""

import oslo_serialization

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common


class Tag(common.CoprHDResource):

    def tag_resource(self, uri, resource_id, add, remove):
        params = {
            'add': add,
            'remove': remove
        }

        body = oslo_serialization.jsonutils.dumps(params)

        (s, h) = common.service_json_request(self.ipaddr, self.port, "PUT",
                                             uri.format(resource_id), body)
        o = common.json_decode(s)
        return o

    def list_tags(self, resource_uri):
        if resource_uri.__contains__("tag") is False:
            raise common.CoprHdError(
                common.CoprHdError.VALUE_ERR, _("URI should end with /tag"))

        (s, h) = common.service_json_request(self.ipaddr,
                                             self.port,
                                             "GET",
                                             resource_uri,
                                             None)

        allTags = []
        o = common.json_decode(s)
        allTags = o['tag']

        return allTags



# Copyright (c) 2013 - 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Driver for Huawei FusionStorage.
"""

import os
import re

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.fusionstorage import fspythonapi

LOG = logging.getLogger(__name__)

volume_opts = [
    cfg.BoolOpt('dsware_isthin',
                default=False,
                help='The flag of thin storage allocation.'),
    cfg.StrOpt('dsware_manager',
               default='',
               help='Fusionstorage manager ip addr for cinder-volume.'),
    cfg.StrOpt('fusionstorageagent',
               default='',
               help='Fusionstorage agent ip addr range.'),
    cfg.StrOpt('pool_type',
               default='default',
               help = 'Pool type, like sata-2copy.'),
    cfg.ListOpt('pool_id_filter',
                default=[],
                help='Pool id permit to use.'),
    cfg.IntOpt('clone_volume_timeout',
               default=680,
               help='Create clone volume timeout.'),
]

CONF = cfg.CONF
CONF.register_opts(volume_opts, group=configuration.SHARED_CONF_GROUP)

OLD_VERSION = 1
NEW_VERSION = 0
VOLUME_ALREADY_ATTACHED = 50151401
VOLUME_NOT_EXIST = '50150005\n'
VOLUME_BEING_DELETED = '50151002\n'
SNAP_NOT_EXIST = '50150006\n'


@interface.volumedriver
class DSWAREDriver(driver.VolumeDriver):
    """Huawei FusionStorage Driver."""
    VERSION = '1.0'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Huawei_FusionStorage_CI"

    DSWARE_VOLUME_CREATE_SUCCESS_STATUS = 0
    DSWARE_VOLUME_DUPLICATE_VOLUME = 6
    DSWARE_VOLUME_CREATING_STATUS = 7

    def __init__(self, *args, **kwargs):
        super(DSWAREDriver, self).__init__(*args, **kwargs)
        self.dsware_client = fspythonapi.FSPythonApi()
        self.check_cloned_interval = 2
        self.configuration.append_config_values(volume_opts)

    def check_for_setup_error(self):
        # lrk: check config file here.
        if not os.path.exists(fspythonapi.fsc_conf_file):
            msg = _("Dsware config file not exists!")
            LOG.error("Dsware config file: %s not exists!",
                      fspythonapi.fsc_conf_file)
            raise exception.VolumeBackendAPIException(data=msg)

    def do_setup(self, context):
        # lrk: create fsc_conf_file here.
        conf_info = ["manage_ip=%s" % self.configuration.dsware_manager,
                     "\n",
                     "vbs_url=%s" % self.configuration.fusionstorageagent]

        fsc_dir = os.path.dirname(fspythonapi.fsc_conf_file)
        if not os.path.exists(fsc_dir):
            os.makedirs(fsc_dir)

        with open(fspythonapi.fsc_conf_file, 'w') as f:
            f.writelines(conf_info)

        # Get pool type.
        self.pool_type = self.configuration.pool_type
        LOG.debug("Dsware Driver do_setup finish.")

    def _get_dsware_manage_ip(self, volume):
        dsw_manager_ip = volume.provider_id
        if dsw_manager_ip is not None:
            return dsw_manager_ip
        else:
            msg = _("Dsware get manager ip failed, "
                    "volume provider_id is None!")
            raise exception.VolumeBackendAPIException(data=msg)

    def _get_poolid_from_host(self, host):
        # Host format: 'hostid@backend#poolid'.
        # Other formats: return 'default', and the pool id would be zero.
        if host:
            if len(host.split('#', 1)) == 2:
                return host.split('#')[1]
        return self.pool_type

    def _create_volume(self, volume_id, volume_size, is_thin, volume_host):
        pool_id = 0
        result = 1

        # Query Dsware version.
        retcode = self.dsware_client.query_dsware_version()
        # Old version.
        if retcode == OLD_VERSION:
            pool_id = 0
        # New version.
        elif retcode == NEW_VERSION:
            pool_info = self._get_poolid_from_host(volume_host)
            if pool_info != self.pool_type:
                pool_id = int(pool_info)
        # Query Dsware version failed!
        else:
            LOG.error("Query Dsware version fail!")
            msg = (_("Query Dsware version failed! Retcode is %s.") %
                   retcode)
            raise exception.VolumeBackendAPIException(data=msg)

        try:
            result = self.dsware_client.create_volume(
                volume_id, pool_id, volume_size, int(is_thin))
        except Exception as e:
            LOG.exception("Create volume error, details is: %s.", e)
            raise

        if result != 0:
            msg = _("Dsware create volume failed! Result is: %s.") % result
            raise exception.VolumeBackendAPIException(data=msg)

    def create_volume(self, volume):
        # Creates a volume in Dsware.
        LOG.debug("Begin to create volume %s in Dsware.", volume.name)
        volume_id = volume.name
        volume_size = volume.size
        volume_host = volume.host
        is_thin = self.configuration.dsware_isthin
        # Change GB to MB.
        volume_size *= 1024
        self._create_volume(volume_id, volume_size, is_thin, volume_host)

        dsw_manager_ip = self.dsware_client.get_manage_ip()
        return {"provider_id": dsw_manager_ip}

    def _create_volume_from_snap(self, volume_id, volume_size, snapshot_name):
        result = self.dsware_client.create_volume_from_snap(
            volume_id, volume_size, snapshot_name)
        if result != 0:
            msg = (_("Dsware: create volume from snap failed. Result: %s.") %
                   result)
            raise exception.VolumeBackendAPIException(data=msg)

    def create_volume_from_snapshot(self, volume, snapshot):
        # Creates a volume from snapshot.
        volume_id = volume.name
        volume_size = volume.size
        snapshot_name = snapshot.name
        if volume_size < int(snapshot.volume_size):
            msg = _("Dsware: volume size can not be less than snapshot size.")
            raise exception.VolumeBackendAPIException(data=msg)
        # Change GB to MB.
        volume_size *= 1024
        self._create_volume_from_snap(volume_id, volume_size, snapshot_name)

        dsw_manager_ip = self.dsware_client.get_manage_ip()
        return {"provider_id": dsw_manager_ip}

    def create_cloned_volume(self, volume, src_volume):
        """Dispatcher to Dsware client to create volume from volume.

        Wait volume create finished.
        """
        volume_name = volume.name
        volume_size = volume.size
        src_volume_name = src_volume.name
        # Change GB to MB.
        volume_size *= 1024
        result = self.dsware_client.create_volume_from_volume(
            volume_name, volume_size, src_volume_name)
        if result:
            msg = _('Dsware fails to start cloning volume %s.') % volume_name
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('Dsware create volume %(volume_name)s of size '
                  '%(volume_size)s from src volume %(src_volume_name)s start.',
                  {"volume_name": volume_name,
                   "volume_size": volume_size,
                   "src_volume_name": src_volume_name})

        ret = self._wait_for_create_cloned_volume_finish_timer(volume_name)
        if not ret:
            msg = (_('Clone volume %s failed while waiting for success.') %
                   volume_name)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('Dsware create volume from volume ends.')

        dsw_manager_ip = self.dsware_client.get_manage_ip()
        return {"provider_id": dsw_manager_ip}

    def _check_create_cloned_volume_finish(self, new_volume_name):
        LOG.debug('Loopcall: _check_create_cloned_volume_finish(), '
                  'volume-name: %s.', new_volume_name)
        current_volume = self.dsware_client.query_volume(new_volume_name)

        if current_volume:
            status = current_volume['status']
            LOG.debug('Wait clone volume %(volume_name)s, status: %(status)s.',
                      {"volume_name": new_volume_name,
                       "status": status})
            if int(status) == self.DSWARE_VOLUME_CREATING_STATUS or int(
                    status) == self.DSWARE_VOLUME_DUPLICATE_VOLUME:
                self.count += 1
            elif int(status) == self.DSWARE_VOLUME_CREATE_SUCCESS_STATUS:
                raise loopingcall.LoopingCallDone(retvalue=True)
            else:
                msg = _('Clone volume %(new_volume_name)s failed, '
                        'volume status is: %(status)s.')
                LOG.error(msg, {'new_volume_name': new_volume_name,
                                'status': status})
                raise loopingcall.LoopingCallDone(retvalue=False)
            if self.count > self.configuration.clone_volume_timeout:
                msg = _('Dsware clone volume time out. '
                        'Volume: %(new_volume_name)s, status: %(status)s')
                LOG.error(msg, {'new_volume_name': new_volume_name,
                                'status': current_volume['status']})
                raise loopingcall.LoopingCallDone(retvalue=False)
        else:
            LOG.warning('Can not find volume %s from Dsware.',
                        new_volume_name)
            self.count += 1
            if self.count > 10:
                msg = _("Dsware clone volume failed: volume "
                        "can not be found from Dsware.")
                LOG.error(msg)
                raise loopingcall.LoopingCallDone(retvalue=False)

    def _wait_for_create_cloned_volume_finish_timer(self, new_volume_name):
        timer = loopingcall.FixedIntervalLoopingCall(
            self._check_create_cloned_volume_finish, new_volume_name)
        LOG.debug('Call _check_create_cloned_volume_finish: volume-name %s.',
                  new_volume_name)
        self.count = 0
        ret = timer.start(interval=self.check_cloned_interval).wait()
        timer.stop()
        return ret

    def _analyse_output(self, out):
        if out is not None:
            analyse_result = {}
            out_temp = out.split('\n')
            for line in out_temp:
                if re.search('^ret_code=', line):
                    analyse_result['ret_code'] = line[9:]
                elif re.search('^ret_desc=', line):
                    analyse_result['ret_desc'] = line[9:]
                elif re.search('^dev_addr=', line):
                    analyse_result['dev_addr'] = line[9:]
            return analyse_result
        else:
            return None

    def _attach_volume(self, volume_name, dsw_manager_ip):
        cmd = ['vbs_cli', '-c', 'attachwithip', '-v', volume_name, '-i',
               dsw_manager_ip.replace('\n', ''), '-p', 0]
        out, err = self._execute(*cmd, run_as_root=True)
        analyse_result = self._analyse_output(out)
        LOG.debug("Attach volume result is %s.", analyse_result)
        return analyse_result

    def _detach_volume(self, volume_name, dsw_manager_ip):
        cmd = ['vbs_cli', '-c', 'detachwithip', '-v', volume_name, '-i',
               dsw_manager_ip.replace('\n', ''), '-p', 0]
        out, err = self._execute(*cmd, run_as_root=True)
        analyse_result = self._analyse_output(out)
        LOG.debug("Detach volume result is %s.", analyse_result)
        return analyse_result

    def _query_volume_attach(self, volume_name, dsw_manager_ip):
        cmd = ['vbs_cli', '-c', 'querydevwithip', '-v', volume_name, '-i',
               dsw_manager_ip.replace('\n', ''), '-p', 0]
        out, err = self._execute(*cmd, run_as_root=True)
        analyse_result = self._analyse_output(out)
        LOG.debug("Query volume attach result is %s.", analyse_result)
        return analyse_result

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        # Copy image to volume.
        # Step1: attach volume to host.
        LOG.debug("Begin to copy image to volume.")
        dsw_manager_ip = self._get_dsware_manage_ip(volume)
        volume_attach_result = self._attach_volume(volume.name,
                                                   dsw_manager_ip)
        volume_attach_path = ''
        if volume_attach_result is not None and int(
                volume_attach_result['ret_code']) == 0:
            volume_attach_path = volume_attach_result['dev_addr']
            LOG.debug("Volume attach path is %s.", volume_attach_path)
        if volume_attach_path == '':
            msg = _("Host attach volume failed!")
            raise exception.VolumeBackendAPIException(data=msg)
            # Step2: fetch the image from image_service and write it to the
            # volume.
        try:
            image_utils.fetch_to_raw(context,
                                     image_service,
                                     image_id,
                                     volume_attach_path,
                                     self.configuration.volume_dd_blocksize)
        finally:
            # Step3: detach volume from host.
            dsw_manager_ip = self._get_dsware_manage_ip(volume)
            volume_detach_result = self._detach_volume(volume.name,
                                                       dsw_manager_ip)
            if volume_detach_result is not None and int(
                    volume_detach_result['ret_code']) != 0:
                msg = (_("Dsware detach volume from host failed: %s!") %
                       volume_detach_result)
                raise exception.VolumeBackendAPIException(data=msg)

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        # Copy volume to image.
        # If volume was not attached, then attach it.

        dsw_manager_ip = self._get_dsware_manage_ip(volume)

        already_attached = False
        _attach_result = self._attach_volume(volume.name, dsw_manager_ip)
        if _attach_result:
            retcode = _attach_result['ret_code']
            if int(retcode) == VOLUME_ALREADY_ATTACHED:
                already_attached = True
                result = self._query_volume_attach(volume.name,
                                                   dsw_manager_ip)
                if not result or int(result['ret_code']) != 0:
                    msg = (_("Query volume attach failed, result=%s.") %
                           result)
                    raise exception.VolumeBackendAPIException(data=msg)

            elif int(retcode) == 0:
                result = _attach_result
            else:
                msg = (_("Attach volume to host failed "
                         "in copy volume to image, retcode: %s.") %
                       retcode)
                raise exception.VolumeBackendAPIException(data=msg)

            volume_attach_path = result['dev_addr']

        else:
            msg = _("Attach_volume failed.")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        try:
            image_utils.upload_volume(context,
                                      image_service,
                                      image_meta,
                                      volume_attach_path)
        except Exception as e:
            LOG.error("Upload volume error, details: %s.", e)
            raise
        finally:
            if not already_attached:
                self._detach_volume(volume.name, dsw_manager_ip)

    def _get_volume(self, volume_name):
        result = self.dsware_client.query_volume(volume_name)
        LOG.debug("Dsware query volume result is %s.", result['result'])
        if result['result'] == VOLUME_NOT_EXIST:
            LOG.debug("Dsware volume %s does not exist.", volume_name)
            return False
        elif result['result'] == 0:
            return True
        else:
            msg = _("Dsware query volume %s failed!") % volume_name
            raise exception.VolumeBackendAPIException(data=msg)

    def _delete_volume(self, volume_name):
        # Delete volume in Dsware.
        result = self.dsware_client.delete_volume(volume_name)
        LOG.debug("Dsware delete volume, result is %s.", result)
        if result == VOLUME_NOT_EXIST:
            LOG.debug("Dsware delete volume, volume does not exist.")
            return True
        elif result == VOLUME_BEING_DELETED:
            LOG.debug("Dsware delete volume, volume is being deleted.")
            return True
        elif result == 0:
            return True
        else:
            msg = _("Dsware delete volume failed: %s!") % result
            raise exception.VolumeBackendAPIException(data=msg)

    def delete_volume(self, volume):
        # Delete volume.
        # If volume does not exist, then return.
        LOG.debug("Begin to delete volume in Dsware: %s.", volume.name)
        if not self._get_volume(volume.name):
            return True

        return self._delete_volume(volume.name)

    def _get_snapshot(self, snapshot_name):
        snapshot_info = self.dsware_client.query_snap(snapshot_name)
        LOG.debug("Get snapshot, snapshot_info is : %s.", snapshot_info)
        if snapshot_info['result'] == SNAP_NOT_EXIST:
            LOG.error('Snapshot: %s not found!', snapshot_name)
            return False
        elif snapshot_info['result'] == 0:
            return True
        else:
            msg = _("Dsware get snapshot failed!")
            raise exception.VolumeBackendAPIException(data=msg)

    def _create_snapshot(self, snapshot_id, volume_id):
        LOG.debug("Create snapshot %s to Dsware.", snapshot_id)
        smart_flag = 0
        res = self.dsware_client.create_snapshot(snapshot_id,
                                                 volume_id,
                                                 smart_flag)
        if res != 0:
            msg = _("Dsware Create Snapshot failed! Result: %s.") % res
            raise exception.VolumeBackendAPIException(data=msg)

    def _delete_snapshot(self, snapshot_id):
        LOG.debug("Delete snapshot %s to Dsware.", snapshot_id)
        res = self.dsware_client.delete_snapshot(snapshot_id)
        LOG.debug("Ddelete snapshot result is: %s.", res)
        if res != 0:
            raise exception.SnapshotIsBusy(snapshot_name=snapshot_id)

    def create_snapshot(self, snapshot):
        vol_id = 'volume-%s' % snapshot.volume_id
        snapshot_id = snapshot.name
        if not self._get_volume(vol_id):
            LOG.error('Create Snapshot, but volume: %s not found!', vol_id)
            raise exception.VolumeNotFound(volume_id=vol_id)
        else:
            self._create_snapshot(snapshot_id, vol_id)

    def delete_snapshot(self, snapshot):
        LOG.debug("Delete snapshot %s.", snapshot.name)
        snapshot_id = snapshot.name
        if self._get_snapshot(snapshot_id):
            self._delete_snapshot(snapshot_id)

    def _calculate_pool_info(self, pool_sets):
        filter = False
        pools_status = []
        reserved_percentage = self.configuration.reserved_percentage
        pool_id_filter = self.configuration.pool_id_filter
        LOG.debug("Filtered pool id is %s.", pool_id_filter)
        if pool_id_filter == []:
            for pool_info in pool_sets:
                pool = {}
                pool['pool_name'] = pool_info['pool_id']
                pool['total_capacity_gb'] = float(
                    pool_info['total_capacity']) / 1024
                pool['allocated_capacity_gb'] = float(
                    pool_info['used_capacity']) / 1024
                pool['free_capacity_gb'] = pool['total_capacity_gb'] - pool[
                    'allocated_capacity_gb']
                pool['QoS_support'] = False
                pool['reserved_percentage'] = reserved_percentage
                pools_status.append(pool)
        else:
            for pool_info in pool_sets:
                for pool_id in pool_id_filter:
                    if pool_id == pool_info['pool_id']:
                        filter = True
                        break

                if filter:
                    pool = {}
                    pool['pool_name'] = pool_info['pool_id']
                    pool['total_capacity_gb'] = float(
                        pool_info['total_capacity']) / 1024
                    pool['allocated_capacity_gb'] = float(
                        pool_info['used_capacity']) / 1024
                    pool['free_capacity_gb'] = float(
                        pool['total_capacity_gb'] - pool[
                            'allocated_capacity_gb'])
                    pool['QoS_support'] = False
                    pool['reserved_percentage'] = reserved_percentage
                    pools_status.append(pool)

                filter = False

        return pools_status

    def _update_single_pool_info_status(self):
        """Query pool info when Dsware is single-pool version."""
        status = {}
        status['volume_backend_name'] = self.configuration.volume_backend_name
        status['vendor_name'] = 'Open Source'
        status['driver_version'] = self.VERSION
        status['storage_protocol'] = 'dsware'

        status['total_capacity_gb'] = 0
        status['free_capacity_gb'] = 0
        status['reserved_percentage'] = self.configuration.reserved_percentage
        status['QoS_support'] = False
        pool_id = 0
        pool_info = self.dsware_client.query_pool_info(pool_id)
        result = pool_info['result']
        if result == 0:
            status['total_capacity_gb'] = float(
                pool_info['total_capacity']) / 1024
            status['free_capacity_gb'] = (float(
                pool_info['total_capacity']) - float(
                pool_info['used_capacity'])) / 1024
            LOG.debug("total_capacity_gb is %s, free_capacity_gb is %s.",
                      status['total_capacity_gb'],
                      status['free_capacity_gb'])
            self._stats = status
        else:
            self._stats = None

    def _update_multi_pool_of_same_type_status(self):
        """Query info of multiple pools when Dsware is multi-pool version.

        These pools have the same pool type.
        """
        status = {}
        status['volume_backend_name'] = self.configuration.volume_backend_name
        status['vendor_name'] = 'Open Source'
        status['driver_version'] = self.VERSION
        status['storage_protocol'] = 'dsware'

        (result, pool_sets) = self.dsware_client.query_pool_type(
            self.pool_type)
        if pool_sets == []:
            self._stats = None
        else:
            pools_status = self._calculate_pool_info(pool_sets)
            status['pools'] = pools_status
            self._stats = status

    def get_volume_stats(self, refresh=False):
        if refresh:
            dsware_version = self.dsware_client.query_dsware_version()
            # Old version.
            if dsware_version == OLD_VERSION:
                self._update_single_pool_info_status()
            # New version.
            elif dsware_version == NEW_VERSION:
                self._update_multi_pool_of_same_type_status()
            else:
                msg = _("Dsware query Dsware version failed!")
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        return self._stats

    def extend_volume(self, volume, new_size):
        # Extend volume in Dsware.
        LOG.debug("Begin to extend volume in Dsware: %s.", volume.name)
        volume_id = volume.name
        if volume.size > new_size:
            msg = (_("Dsware extend Volume failed! "
                     "New size %(new_size)s should be greater than "
                     "old size %(old_size)s!")
                   % {'new_size': new_size,
                      'old_size': volume.size})
            raise exception.VolumeBackendAPIException(data=msg)
        # Change GB to MB.
        volume_size = new_size * 1024
        result = self.dsware_client.extend_volume(volume_id, volume_size)
        if result != 0:
            msg = _("Dsware extend Volume failed! Result:%s.") % result
            raise exception.VolumeBackendAPIException(data=msg)

    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info."""
        LOG.debug("Begin initialize connection.")

        properties = {}
        properties['volume_name'] = volume.name
        properties['volume'] = volume
        properties['dsw_manager_ip'] = self._get_dsware_manage_ip(volume)

        LOG.debug("End initialize connection with properties:%s.", properties)

        return {'driver_volume_type': 'dsware',
                'data': properties}

    def terminate_connection(self, volume, connector, force=False, **kwargs):
        pass

    def create_export(self, context, volume, connector):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass



# (c) Copyright 2015 Industrial Technology Research Institute.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test case for the delete volume function."""

import mock

from cinder import exception
from cinder.tests.unit.volume.drivers import disco


class DeleteVolumeTestCase(disco.TestDISCODriver):
    """Test cases to delete DISCO volumes."""

    def setUp(self):
        """Initialise variables and mock functions."""
        super(DeleteVolumeTestCase, self).setUp()

        # Mock volumeDelete function.
        mock.patch.object(self.requester,
                          'volumeDelete',
                          self.perform_disco_request).start()

        self.response = self.FAKE_RESPONSE['standard']['success']

    def perform_disco_request(self, *cmd, **kwargs):
        """Mock function to delete a volume."""
        return self.response

    def test_delete_volume(self):
        """Delete a volume."""
        self.driver.delete_volume(self.volume)

    def test_delete_volume_fail(self):
        """Make the API returns an error while deleting."""
        self.response = self.FAKE_RESPONSE['standard']['fail']
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.test_delete_volume)



# Copyright (c) 2016 EMC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The groups controller."""

from oslo_log import log as logging
from oslo_utils import strutils
from oslo_utils import uuidutils
from six.moves import http_client
import webob
from webob import exc

from cinder.api import common
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.v3.views import groups as views_groups
from cinder import exception
from cinder import group as group_api
from cinder.i18n import _
from cinder import rpc
from cinder.volume import group_types

LOG = logging.getLogger(__name__)


class GroupsController(wsgi.Controller):
    """The groups API controller for the OpenStack API."""

    _view_builder_class = views_groups.ViewBuilder

    def __init__(self):
        self.group_api = group_api.API()
        super(GroupsController, self).__init__()

    def _check_default_cgsnapshot_type(self, group_type_id):
        if group_types.is_default_cgsnapshot_type(group_type_id):
            msg = _("Group_type %(group_type)s is reserved for migrating "
                    "CGs to groups. Migrated group can only be operated by "
                    "CG APIs.") % {'group_type': group_type_id}
            raise exc.HTTPBadRequest(explanation=msg)

    @wsgi.Controller.api_version(mv.GROUP_VOLUME)
    def show(self, req, id):
        """Return data about the given group."""
        LOG.debug('show called for member %s', id)
        context = req.environ['cinder.context']

        # Not found exception will be handled at the wsgi level
        group = self.group_api.get(
            context,
            group_id=id)

        self._check_default_cgsnapshot_type(group.group_type_id)

        return self._view_builder.detail(req, group)

    @wsgi.Controller.api_version(mv.GROUP_VOLUME_RESET_STATUS)
    @wsgi.action("reset_status")
    def reset_status(self, req, id, body):
        return self._reset_status(req, id, body)

    def _reset_status(self, req, id, body):
        """Reset status on generic group."""

        context = req.environ['cinder.context']
        try:
            status = body['reset_status']['status'].lower()
        except (TypeError, KeyError):
            raise exc.HTTPBadRequest(explanation=_("Must specify 'status'"))

        LOG.debug("Updating group '%(id)s' with "
                  "'%(update)s'", {'id': id,
                                   'update': status})
        try:
            notifier = rpc.get_notifier('groupStatusUpdate')
            notifier.info(context, 'groups.reset_status.start',
                          {'id': id,
                           'update': status})
            group = self.group_api.get(context, id)

            self.group_api.reset_status(context, group, status)
            notifier.info(context, 'groups.reset_status.end',
                          {'id': id,
                           'update': status})
        except exception.GroupNotFound as error:
            # Not found exception will be handled at the wsgi level
            notifier.error(context, 'groups.reset_status',
                           {'error_message': error.msg,
                            'id': id})
            raise
        except exception.InvalidGroupStatus as error:
            notifier.error(context, 'groups.reset_status',
                           {'error_message': error.msg,
                            'id': id})
            raise exc.HTTPBadRequest(explanation=error.msg)
        return webob.Response(status_int=http_client.ACCEPTED)

    @wsgi.Controller.api_version(mv.GROUP_VOLUME)
    @wsgi.action("delete")
    def delete_group(self, req, id, body):
        return self._delete(req, id, body)

    def _delete(self, req, id, body):
        """Delete a group."""
        LOG.debug('delete called for group %s', id)
        context = req.environ['cinder.context']
        del_vol = False
        if body:
            self.assert_valid_body(body, 'delete')

            grp_body = body['delete']
            try:
                del_vol = strutils.bool_from_string(
                    grp_body.get('delete-volumes', False),
                    strict=True)
            except ValueError:
                msg = (_("Invalid value '%s' for delete-volumes flag.")
                       % del_vol)
                raise exc.HTTPBadRequest(explanation=msg)

        LOG.info('Delete group with id: %s', id,
                 context=context)

        try:
            group = self.group_api.get(context, id)
            self._check_default_cgsnapshot_type(group.group_type_id)
            self.group_api.delete(context, group, del_vol)
        except exception.GroupNotFound:
            # Not found exception will be handled at the wsgi level
            raise
        except exception.InvalidGroup as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        return webob.Response(status_int=http_client.ACCEPTED)

    @wsgi.Controller.api_version(mv.GROUP_VOLUME)
    def index(self, req):
        """Returns a summary list of groups."""
        return self._get_groups(req, is_detail=False)

    @wsgi.Controller.api_version(mv.GROUP_VOLUME)
    def detail(self, req):
        """Returns a detailed list of groups."""
        return self._get_groups(req, is_detail=True)

    def _get_groups(self, req, is_detail):
        """Returns a list of groups through view builder."""
        context = req.environ['cinder.context']
        filters = req.params.copy()
        api_version = req.api_version_request
        marker, limit, offset = common.get_pagination_params(filters)
        sort_keys, sort_dirs = common.get_sort_params(filters)

        filters.pop('list_volume', None)
        if api_version.matches(mv.RESOURCE_FILTER):
            support_like = (True if api_version.matches(
                mv.LIKE_FILTER) else False)
            common.reject_invalid_filters(context, filters, 'group',
                                          support_like)

        groups = self.group_api.get_all(
            context, filters=filters, marker=marker, limit=limit,
            offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)

        new_groups = []
        for grp in groups:
            try:
                # Only show groups not migrated from CGs
                self._check_default_cgsnapshot_type(grp.group_type_id)
                new_groups.append(grp)
            except exc.HTTPBadRequest:
                # Skip migrated group
                pass

        if is_detail:
            groups = self._view_builder.detail_list(
                req, new_groups)
        else:
            groups = self._view_builder.summary_list(
                req, new_groups)
        return groups

    @wsgi.Controller.api_version(mv.GROUP_VOLUME)
    @wsgi.response(http_client.ACCEPTED)
    def create(self, req, body):
        """Create a new group."""
        LOG.debug('Creating new group %s', body)
        self.assert_valid_body(body, 'group')

        context = req.environ['cinder.context']
        group = body['group']
        self.validate_name_and_description(group)
        name = group.get('name')
        description = group.get('description')
        group_type = group.get('group_type')
        if not group_type:
            msg = _("group_type must be provided to create "
                    "group %(name)s.") % {'name': name}
            raise exc.HTTPBadRequest(explanation=msg)
        if not uuidutils.is_uuid_like(group_type):
            req_group_type = group_types.get_group_type_by_name(context,
                                                                group_type)
            group_type = req_group_type['id']
        self._check_default_cgsnapshot_type(group_type)
        volume_types = group.get('volume_types')
        if not volume_types:
            msg = _("volume_types must be provided to create "
                    "group %(name)s.") % {'name': name}
            raise exc.HTTPBadRequest(explanation=msg)
        availability_zone = group.get('availability_zone')

        LOG.info("Creating group %(name)s.",
                 {'name': name},
                 context=context)

        try:
            new_group = self.group_api.create(
                context, name, description, group_type, volume_types,
                availability_zone=availability_zone)
        except (exception.Invalid, exception.ObjectActionError) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        except exception.NotFound:
            # Not found exception will be handled at the wsgi level
            raise

        retval = self._view_builder.summary(req, new_group)
        return retval

    @wsgi.Controller.api_version(mv.GROUP_SNAPSHOTS)
    @wsgi.action("create-from-src")
    @wsgi.response(http_client.ACCEPTED)
    def create_from_src(self, req, body):
        """Create a new group from a source.

        The source can be a group snapshot or a group. Note that
        this does not require group_type and volume_types as the
        "create" API above.
        """
        LOG.debug('Creating new group %s.', body)
        self.assert_valid_body(body, 'create-from-src')

        context = req.environ['cinder.context']
        group = body['create-from-src']
        self.validate_name_and_description(group)
        name = group.get('name', None)
        description = group.get('description', None)
        group_snapshot_id = group.get('group_snapshot_id', None)
        source_group_id = group.get('source_group_id', None)
        if not group_snapshot_id and not source_group_id:
            msg = (_("Either 'group_snapshot_id' or 'source_group_id' must be "
                     "provided to create group %(name)s from source.")
                   % {'name': name})
            raise exc.HTTPBadRequest(explanation=msg)

        if group_snapshot_id and source_group_id:
            msg = _("Cannot provide both 'group_snapshot_id' and "
                    "'source_group_id' to create group %(name)s from "
                    "source.") % {'name': name}
            raise exc.HTTPBadRequest(explanation=msg)

        group_type_id = None
        if group_snapshot_id:
            LOG.info("Creating group %(name)s from group_snapshot "
                     "%(snap)s.",
                     {'name': name, 'snap': group_snapshot_id},
                     context=context)
            grp_snap = self.group_api.get_group_snapshot(context,
                                                         group_snapshot_id)
            group_type_id = grp_snap.group_type_id
        elif source_group_id:
            LOG.info("Creating group %(name)s from "
                     "source group %(source_group_id)s.",
                     {'name': name, 'source_group_id': source_group_id},
                     context=context)
            source_group = self.group_api.get(context, source_group_id)
            group_type_id = source_group.group_type_id

        self._check_default_cgsnapshot_type(group_type_id)

        try:
            new_group = self.group_api.create_from_src(
                context, name, description, group_snapshot_id, source_group_id)
        except exception.InvalidGroup as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        except (exception.GroupNotFound, exception.GroupSnapshotNotFound):
            # Not found exception will be handled at the wsgi level
            raise
        except exception.CinderException as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        retval = self._view_builder.summary(req, new_group)
        return retval

    @wsgi.Controller.api_version(mv.GROUP_VOLUME)
    def update(self, req, id, body):
        """Update the group.

        Expected format of the input parameter 'body':

        .. code-block:: json

            {
                "group":
                {
                    "name": "my_group",
                    "description": "My group",
                    "add_volumes": "volume-uuid-1,volume-uuid-2,...",
                    "remove_volumes": "volume-uuid-8,volume-uuid-9,..."
                }
            }

        """
        LOG.debug('Update called for group %s.', id)

        if not body:
            msg = _("Missing request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        self.assert_valid_body(body, 'group')
        context = req.environ['cinder.context']

        group = body.get('group')
        self.validate_name_and_description(group)
        name = group.get('name')
        description = group.get('description')
        add_volumes = group.get('add_volumes')
        remove_volumes = group.get('remove_volumes')

        # Allow name or description to be changed to an empty string ''.
        if (name is None and description is None and not add_volumes
                and not remove_volumes):
            msg = _("Name, description, add_volumes, and remove_volumes "
                    "can not be all empty in the request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        LOG.info("Updating group %(id)s with name %(name)s "
                 "description: %(description)s add_volumes: "
                 "%(add_volumes)s remove_volumes: %(remove_volumes)s.",
                 {'id': id, 'name': name,
                  'description': description,
                  'add_volumes': add_volumes,
                  'remove_volumes': remove_volumes},
                 context=context)

        try:
            group = self.group_api.get(context, id)
            self._check_default_cgsnapshot_type(group.group_type_id)
            self.group_api.update(
                context, group, name, description,
                add_volumes, remove_volumes)
        except exception.GroupNotFound:
            # Not found exception will be handled at the wsgi level
            raise
        except exception.InvalidGroup as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        return webob.Response(status_int=http_client.ACCEPTED)

    @wsgi.Controller.api_version(mv.GROUP_REPLICATION)
    @wsgi.action("enable_replication")
    def enable_replication(self, req, id, body):
        """Enables replications for a group."""
        context = req.environ['cinder.context']
        if body:
            self.assert_valid_body(body, 'enable_replication')

        LOG.info('Enable replication group with id: %s.', id,
                 context=context)

        try:
            group = self.group_api.get(context, id)
            self.group_api.enable_replication(context, group)
            # Not found exception will be handled at the wsgi level
        except (exception.InvalidGroup, exception.InvalidGroupType,
                exception.InvalidVolume, exception.InvalidVolumeType) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        return webob.Response(status_int=202)

    @wsgi.Controller.api_version(mv.GROUP_REPLICATION)
    @wsgi.action("disable_replication")
    def disable_replication(self, req, id, body):
        """Disables replications for a group."""
        context = req.environ['cinder.context']
        if body:
            self.assert_valid_body(body, 'disable_replication')

        LOG.info('Disable replication group with id: %s.', id,
                 context=context)

        try:
            group = self.group_api.get(context, id)
            self.group_api.disable_replication(context, group)
            # Not found exception will be handled at the wsgi level
        except (exception.InvalidGroup, exception.InvalidGroupType,
                exception.InvalidVolume, exception.InvalidVolumeType) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        return webob.Response(status_int=202)

    @wsgi.Controller.api_version(mv.GROUP_REPLICATION)
    @wsgi.action("failover_replication")
    def failover_replication(self, req, id, body):
        """Fails over replications for a group."""
        context = req.environ['cinder.context']
        if body:
            self.assert_valid_body(body, 'failover_replication')

            grp_body = body['failover_replication']
            try:
                allow_attached = strutils.bool_from_string(
                    grp_body.get('allow_attached_volume', False),
                    strict=True)
            except ValueError:
                msg = (_("Invalid value '%s' for allow_attached_volume flag.")
                       % grp_body)
                raise exc.HTTPBadRequest(explanation=msg)
            secondary_backend_id = grp_body.get('secondary_backend_id')

        LOG.info('Failover replication group with id: %s.', id,
                 context=context)

        try:
            group = self.group_api.get(context, id)
            self.group_api.failover_replication(context, group, allow_attached,
                                                secondary_backend_id)
            # Not found exception will be handled at the wsgi level
        except (exception.InvalidGroup, exception.InvalidGroupType,
                exception.InvalidVolume, exception.InvalidVolumeType) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        return webob.Response(status_int=202)

    @wsgi.Controller.api_version(mv.GROUP_REPLICATION)
    @wsgi.action("list_replication_targets")
    def list_replication_targets(self, req, id, body):
        """List replication targets for a group."""
        context = req.environ['cinder.context']
        if body:
            self.assert_valid_body(body, 'list_replication_targets')

        LOG.info('List replication targets for group with id: %s.', id,
                 context=context)

        # Not found exception will be handled at the wsgi level
        group = self.group_api.get(context, id)
        replication_targets = self.group_api.list_replication_targets(
            context, group)

        return replication_targets


def create_resource():
    return wsgi.Resource(GroupsController())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer
from sqlalchemy import MetaData, String, Table


def upgrade(migrate_engine):
    """Add attachment_specs table."""

    meta = MetaData()
    meta.bind = migrate_engine
    Table('volume_attachment', meta, autoload=True)

    attachment_specs = Table(
        'attachment_specs', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(), default=False),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('attachment_id', String(36),
               ForeignKey('volume_attachment.id'),
               nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    attachment_specs.create()



# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from cinder.tests.functional import functional_helpers
from cinder.volume import configuration


class VolumesTest(functional_helpers._FunctionalTestBase):
    _vol_type_name = 'functional_test_type'

    def setUp(self):
        super(VolumesTest, self).setUp()
        self.api.create_type(self._vol_type_name)

    def _get_flags(self):
        f = super(VolumesTest, self)._get_flags()
        f['volume_driver'] = (
            {'v': 'cinder.tests.fake_driver.FakeLoggingVolumeDriver',
             'g': configuration.SHARED_CONF_GROUP})
        f['default_volume_type'] = {'v': self._vol_type_name}
        return f

    def test_get_volumes_summary(self):
        """Simple check that listing volumes works."""
        volumes = self.api.get_volumes(False)
        self.assertIsNotNone(volumes)

    def test_get_volumes(self):
        """Simple check that listing volumes works."""
        volumes = self.api.get_volumes()
        self.assertIsNotNone(volumes)

    def test_create_and_delete_volume(self):
        """Creates and deletes a volume."""

        # Create volume
        created_volume = self.api.post_volume({'volume': {'size': 1}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(self._vol_type_name, found_volume['volume_type'])

        # It should also be in the all-volume list
        volumes = self.api.get_volumes()
        volume_names = [volume['id'] for volume in volumes]
        self.assertIn(created_volume_id, volume_names)

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])

        # It should be available...
        self.assertEqual('available', found_volume['status'])

        # Delete the volume
        self.api.delete_volume(created_volume_id)

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])

        # Should be gone
        self.assertIsNone(found_volume)

    def test_create_volume_with_metadata(self):
        """Creates a volume with metadata."""

        # Create volume
        metadata = {'key1': 'value1',
                    'key2': 'value2'}
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'metadata': metadata}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there and metadata present
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(metadata, found_volume['metadata'])

    def test_create_volume_in_availability_zone(self):
        """Creates a volume in availability_zone."""

        # Create volume
        availability_zone = 'nova'
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'availability_zone': availability_zone}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there and availability zone present
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(availability_zone, found_volume['availability_zone'])

    def test_create_and_update_volume(self):
        # Create vol1
        created_volume = self.api.post_volume({'volume': {
            'size': 1, 'name': 'vol1'}})
        self.assertEqual('vol1', created_volume['name'])
        created_volume_id = created_volume['id']

        # update volume
        body = {'volume': {'name': 'vol-one'}}
        updated_volume = self.api.put_volume(created_volume_id, body)
        self.assertEqual('vol-one', updated_volume['name'])

        # check for update
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual('vol-one', found_volume['name'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import utils
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer
from sqlalchemy import MetaData, String, Table


def upgrade(migrate_engine):
    """Add backup_metadata table."""

    meta = MetaData()
    meta.bind = migrate_engine

    Table('backups', meta, autoload=True)

    backup_metadata = Table(
        'backup_metadata', meta,
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', Boolean(), default=False),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('backup_id', String(36),
               ForeignKey('backups.id'),
               nullable=False),
        Column('key', String(255)),
        Column('value', String(255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    backup_metadata.create()

    if not utils.index_exists_on_columns(migrate_engine,
                                         'backup_metadata',
                                         ['backup_id']):
        utils.add_index(migrate_engine,
                        'backup_metadata',
                        'backup_metadata_backup_id_idx',
                        ['backup_id'])



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_serialization
from oslo_utils import units
import six

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common
from cinder.volume.drivers.coprhd.helpers import consistencygroup
from cinder.volume.drivers.coprhd.helpers import project
from cinder.volume.drivers.coprhd.helpers import virtualarray
from cinder.volume.drivers.coprhd.helpers import virtualpool


class Volume(common.CoprHDResource):

    # Commonly used URIs for the 'Volume' module
    URI_SEARCH_VOLUMES = '/block/volumes/search?project={0}'
    URI_SEARCH_VOLUMES_BY_TAG = '/block/volumes/search?tag={0}'
    URI_VOLUMES = '/block/volumes'
    URI_VOLUME = URI_VOLUMES + '/{0}'
    URI_VOLUME_EXPORTS = URI_VOLUME + '/exports'
    URI_BULK_DELETE = URI_VOLUMES + '/deactivate'
    URI_DEACTIVATE = URI_VOLUME + '/deactivate'
    URI_EXPAND = URI_VOLUME + '/expand'
    URI_TAG_VOLUME = URI_VOLUME + "/tags"
    URI_VOLUME_CHANGE_VPOOL = URI_VOLUMES + "/vpool-change"

    # Protection REST APIs - clone
    URI_VOLUME_PROTECTION_FULLCOPIES = (
        '/block/volumes/{0}/protection/full-copies')
    URI_SNAPSHOT_PROTECTION_FULLCOPIES = (
        '/block/snapshots/{0}/protection/full-copies')

    URI_VOLUME_CLONE_DETACH = "/block/full-copies/{0}/detach"

    # New CG URIs
    URI_CG_CLONE = "/block/consistency-groups/{0}/protection/full-copies"
    URI_CG_CLONE_DETACH = (
        "/block/consistency-groups/{0}/protection/full-copies/{1}/detach")

    VOLUMES = 'volumes'
    CG = 'consistency-groups'
    BLOCK = 'block'
    SNAPSHOTS = 'snapshots'

    # Lists volumes in a project
    def list_volumes(self, project):
        """Makes REST API call to list volumes under a project.

        :param project: name of project
        :returns: List of volumes uuids in JSON response payload
        """

        volume_uris = self.search_volumes(project)
        volumes = []
        for uri in volume_uris:
            volume = self.show_by_uri(uri)
            if volume:
                volumes.append(volume)
        return volumes

    def search_volumes(self, project_name):

        proj = project.Project(self.ipaddr, self.port)
        project_uri = proj.project_query(project_name)

        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "GET",
                                             Volume.URI_SEARCH_VOLUMES.format(
                                                 project_uri),
                                             None)
        o = common.json_decode(s)
        if not o:
            return []

        volume_uris = []
        resources = common.get_node_value(o, "resource")
        for resource in resources:
            volume_uris.append(resource["id"])
        return volume_uris

    # Shows volume information given its uri
    def show_by_uri(self, uri):
        """Makes REST API call and retrieves volume details based on UUID.

        :param uri: UUID of volume
        :returns: Volume details in JSON response payload
        """

        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "GET",
                                             Volume.URI_VOLUME.format(uri),
                                             None)
        o = common.json_decode(s)
        inactive = common.get_node_value(o, 'inactive')
        if inactive:
            return None
        return o

    # Creates a volume given label, project, vpool and size
    def create(self, project_name, label, size, varray, vpool,
               sync, consistencygroup, synctimeout=0):
        """Makes REST API call to create volume under a project.

        :param project_name: name of the project under which the volume
                             will be created
        :param label: name of volume
        :param size: size of volume
        :param varray: name of varray
        :param vpool: name of vpool
        :param sync: synchronous request
        :param consistencygroup: To create volume under a consistencygroup
        :param synctimeout: Query for task status for 'synctimeout' secs.
                            If the task doesn't complete in synctimeout secs,
                            an exception is thrown
        :returns: Created task details in JSON response payload
        """

        proj_obj = project.Project(self.ipaddr, self.port)
        project_uri = proj_obj.project_query(project_name)

        vpool_obj = virtualpool.VirtualPool(self.ipaddr, self.port)
        vpool_uri = vpool_obj.vpool_query(vpool, "block")

        varray_obj = virtualarray.VirtualArray(self.ipaddr, self.port)
        varray_uri = varray_obj.varray_query(varray)

        request = {
            'name': label,
            'size': size,
            'varray': varray_uri,
            'project': project_uri,
            'vpool': vpool_uri,
            'count': 1
        }
        if consistencygroup:
            request['consistency_group'] = consistencygroup

        body = oslo_serialization.jsonutils.dumps(request)
        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "POST",
                                             Volume.URI_VOLUMES,
                                             body)
        o = common.json_decode(s)

        if sync:
            # check task empty
            if len(o["task"]) > 0:
                task = o["task"][0]
                return self.check_for_sync(task, sync, synctimeout)
            else:
                raise common.CoprHdError(
                    common.CoprHdError.SOS_FAILURE_ERR,
                    _("error: task list is empty, no task response found"))
        else:
            return o

    # Blocks the operation until the task is complete/error out/timeout
    def check_for_sync(self, result, sync, synctimeout=0):
        if sync:
            if len(result["resource"]) > 0:
                resource = result["resource"]
                return (
                    common.block_until_complete("volume", resource["id"],
                                                result["id"], self.ipaddr,
                                                self.port, synctimeout)
                )
            else:
                raise common.CoprHdError(
                    common.CoprHdError.SOS_FAILURE_ERR,
                    _("error: task list is empty, no task response found"))
        else:
            return result

    # Queries a volume given its name
    def volume_query(self, full_project_name, volume_name):
        """Makes REST API call to query the volume by name.

        :param volume_name: name of volume
        :param full_project_name: Full project path
        :returns: Volume details in JSON response payload
        """
        if common.is_uri(volume_name):
            return volume_name

        if not full_project_name:
            raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                     _("Project name not specified"))
        uris = self.search_volumes(full_project_name)
        for uri in uris:
            volume = self.show_by_uri(uri)
            if volume and 'name' in volume and volume['name'] == volume_name:
                return volume['id']
        raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                 (_("Volume"
                                    "%s: not found") % volume_name))

    def get_storageAttributes(self, volume_name, cg_name, snapshot_name=None):
        storageres_type = None
        storageres_typename = None

        if snapshot_name is not None:
            storageres_type = Volume.BLOCK
            storageres_typename = Volume.SNAPSHOTS
        elif volume_name is not None:
            storageres_type = Volume.BLOCK
            storageres_typename = Volume.VOLUMES
        elif cg_name is not None:
            storageres_type = Volume.BLOCK
            storageres_typename = Volume.CG
        else:
            storageres_type = None
            storageres_typename = None
        return (storageres_type, storageres_typename)

    def storage_resource_query(self,
                               storageres_type,
                               volume_name,
                               cg_name,
                               snapshot_name,
                               project,
                               tenant):
        resourcepath = "/" + project
        if tenant is not None:
            resourcepath = tenant + resourcepath

        resUri = None
        resourceObj = None

        if Volume.BLOCK == storageres_type and volume_name is not None:
            resUri = self.volume_query(resourcepath, volume_name)
            if snapshot_name is not None:
                from cinder.volume.drivers.coprhd.helpers import snapshot
                snapobj = snapshot.Snapshot(self.ipaddr, self.port)
                resUri = snapobj.snapshot_query(storageres_type,
                                                Volume.VOLUMES, resUri,
                                                snapshot_name)

        elif Volume.BLOCK == storageres_type and cg_name is not None:
            resourceObj = consistencygroup.ConsistencyGroup(
                self.ipaddr, self.port)
            resUri = resourceObj.consistencygroup_query(
                cg_name,
                project,
                tenant)
        else:
            resourceObj = None

        return resUri

    # Creates volume(s) from given source volume
    def clone(self, new_vol_name, resource_uri,
              sync, synctimeout=0):
        """Makes REST API call to clone volume.

        :param new_vol_name: name of volume
        :param resource_uri: uri of source volume
        :param sync: synchronous request
        :param synctimeout: Query for task status for 'synctimeout' secs.
                            If the task doesn't complete in synctimeout secs,
                            an exception is thrown
        :returns: Created task details in JSON response payload
        """
        is_snapshot_clone = False
        clone_full_uri = None

        # consistency group
        if resource_uri.find("BlockConsistencyGroup") > 0:
            clone_full_uri = Volume.URI_CG_CLONE.format(resource_uri)
        elif resource_uri.find("BlockSnapshot") > 0:
            is_snapshot_clone = True
            clone_full_uri = (
                Volume.URI_SNAPSHOT_PROTECTION_FULLCOPIES.format(resource_uri))
        else:
            clone_full_uri = (
                Volume.URI_VOLUME_PROTECTION_FULLCOPIES.format(resource_uri))

        request = {
            'name': new_vol_name,
            'type': None,
            'count': 1
        }

        request["count"] = 1

        body = oslo_serialization.jsonutils.dumps(request)
        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "POST",
                                             clone_full_uri,
                                             body)
        o = common.json_decode(s)

        if sync:
            task = o["task"][0]

            if is_snapshot_clone:
                return (
                    common.block_until_complete(
                        "block",
                        task["resource"]["id"],
                        task["id"], self.ipaddr, self.port)
                )
            else:
                return self.check_for_sync(task, sync, synctimeout)
        else:
            return o

    # To check whether a cloned volume is in detachable state or not
    def is_volume_detachable(self, full_project_name, name):

        volume_uri = self.volume_query(full_project_name, name)
        vol = self.show_by_uri(volume_uri)
        # Filtering based on "replicaState" attribute value of Cloned volume.
        # If "replicaState" value is "SYNCHRONIZED" then only Cloned volume
        # would be in detachable state.
        try:
            return vol['protection']['full_copies'][
                'replicaState'] == 'SYNCHRONIZED'
        except TypeError:
            return False

    def volume_clone_detach(self, resource_uri, full_project_name,
                            name, sync, synctimeout=0):

        volume_uri = self.volume_query(full_project_name, name)

        # consistency group
        if resource_uri.find("BlockConsistencyGroup") > 0:
            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "POST",
                Volume.URI_CG_CLONE_DETACH.format(
                    resource_uri,
                    volume_uri), None)
        else:
            (s, h) = common.service_json_request(
                self.ipaddr, self.port,
                "POST",
                Volume.URI_VOLUME_CLONE_DETACH.format(volume_uri), None)

        o = common.json_decode(s)
        if sync:
            task = o["task"][0]
            return self.check_for_sync(task, sync, synctimeout)
        else:
            return o

    # Shows volume information given its name
    def show(self, full_project_name, name):
        """Retrieves volume details based on volume name.

        :param full_project_name: project path of the volume
        :param name: name of the volume. If the volume is under a project,
                     then full XPath needs to be specified.
                     Example: If VOL1 is a volume under project PROJ1,
                     then the name of volume is PROJ1/VOL1
        :returns: Volume details in JSON response payload
        """
        if common.is_uri(name):
            return name
        if full_project_name is None:
            raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                     (_("Volume %s : not found") %
                                      six.text_type(name)))

        uris = self.search_volumes(full_project_name)

        for uri in uris:
            volume = self.show_by_uri(uri)
            if volume and 'name' in volume and volume['name'] == name:
                return volume
        raise common.CoprHdError(common.CoprHdError.NOT_FOUND_ERR,
                                 (_("Volume"
                                    " %s : not found") % six.text_type(name)))

    def expand(self, full_project_name, volume_name, new_size,
               sync=False, synctimeout=0):

        volume_detail = self.show(full_project_name, volume_name)
        from decimal import Decimal
        new_size_in_gb = Decimal(Decimal(new_size) / (units.Gi))
        current_size = Decimal(volume_detail["provisioned_capacity_gb"])
        if new_size_in_gb <= current_size:
            raise common.CoprHdError(
                common.CoprHdError.VALUE_ERR,
                (_("error: Incorrect value of new size: %(new_size_in_gb)s"
                   " GB\nNew size must be greater than current size: "
                   "%(current_size)s GB") % {'new_size_in_gb': new_size_in_gb,
                                             'current_size': current_size}))

        body = oslo_serialization.jsonutils.dumps({
            "new_size": new_size
        })

        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "POST",
                                             Volume.URI_EXPAND.format(
                                                 volume_detail["id"]),
                                             body)
        if not s:
            return None
        o = common.json_decode(s)

        if sync:
            return self.check_for_sync(o, sync, synctimeout)
        return o

    # Deletes a volume given a volume name
    def delete(self, full_project_name, name, sync=False,
               force_delete=False, coprhdonly=False, synctimeout=0):
        """Deletes a volume based on volume name.

        :param full_project_name: project name
        :param name: name of volume to be deleted
        :param sync: synchronous request
        :param force_delete: if true, it will force the delete of internal
                             volumes that have the SUPPORTS_FORCE flag
        :param coprhdonly: to delete volumes from coprHD only
        :param synctimeout: Query for task status for 'synctimeout' secs. If
                            the task doesn't complete in synctimeout secs,
                            an exception is thrown
        """
        volume_uri = self.volume_query(full_project_name, name)
        return self.delete_by_uri(volume_uri, sync, force_delete,
                                  coprhdonly, synctimeout)

    # Deletes a volume given a volume uri
    def delete_by_uri(self, uri, sync=False,
                      force_delete=False, coprhdonly=False, synctimeout=0):
        """Deletes a volume based on volume uri."""
        params = ''
        if force_delete:
            params += '&' if ('?' in params) else '?'
            params += "force=" + "true"
        if coprhdonly is True:
            params += '&' if ('?' in params) else '?'
            params += "type=" + 'CoprHD_ONLY'

        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "POST",
                                             Volume.URI_DEACTIVATE.format(
                                                 uri) + params,
                                             None)
        if not s:
            return None
        o = common.json_decode(s)
        if sync:
            return self.check_for_sync(o, sync, synctimeout)
        return o

    # Gets the exports info given a volume uri
    def get_exports_by_uri(self, uri):
        """Makes REST API call to get exports info of a volume.

        :param uri: URI of the volume
        :returns: Exports details in JSON response payload
        """
        (s, h) = common.service_json_request(self.ipaddr, self.port,
                                             "GET",
                                             Volume.URI_VOLUME_EXPORTS.format(
                                                 uri),
                                             None)
        return common.json_decode(s)

    # Update a volume information
    # Changed the volume vpool
    def update(self, prefix_path, name, vpool):
        """Makes REST API call to update a volume information.

        :param name: name of the volume to be updated
        :param vpool: name of vpool
        :returns: Created task details in JSON response payload
        """
        namelist = []

        if isinstance(name, list):
            namelist = name
        else:
            namelist.append(name)

        volumeurilist = []

        for item in namelist:
            volume_uri = self.volume_query(prefix_path, item)
            volumeurilist.append(volume_uri)

        vpool_obj = virtualpool.VirtualPool(self.ipaddr, self.port)
        vpool_uri = vpool_obj.vpool_query(vpool, "block")

        params = {
            'vpool': vpool_uri,
            'volumes': volumeurilist
        }

        body = oslo_serialization.jsonutils.dumps(params)

        (s, h) = common.service_json_request(
            self.ipaddr, self.port, "POST",
            Volume.URI_VOLUME_CHANGE_VPOOL,
            body)

        o = common.json_decode(s)
        return o
